"""This module defines a serializer for the transaction model."""
from decimal import Decimal
from datetime import timezone

from rest_framework import serializers
from django.db.models import QuerySet

from polaris import settings
from polaris.models import Transaction
from polaris.settings import DATETIME_FORMAT


class TransactionSerializer(serializers.ModelSerializer):
    """Defines the custom serializer for a transaction."""

    id = serializers.CharField()
    message = serializers.CharField()
    started_at = serializers.DateTimeField(
        format=DATETIME_FORMAT, default_timezone=timezone.utc
    )
    completed_at = serializers.DateTimeField(
        format=DATETIME_FORMAT, default_timezone=timezone.utc
    )

    def __init__(self, data, *args, **kwargs):
        """
        Saves the transaction's asset to the object instance so _round_decimals()
        does not have to access instance.asset, making a new DB query each time.

        This class asssumes the transaction objects being passed use the same
        asset. If the transactions to be serialized are for multiple assets,
        split the calls to this serializer by asset.
        """
        if isinstance(data, Transaction):
            self.asset = data.asset
        elif isinstance(data, list) or isinstance(data, QuerySet):
            if isinstance(data, QuerySet):
                data = list(data)
            if len(data) != 0:
                self.asset = data[0].asset
        super().__init__(data, *args, **kwargs)

    def to_representation(self, instance):
        """
        Removes the "address" part of the to_address and from_address field
        names from the serialized representation. Also removes the
        deposit-related fields for withdraw transactions and vice-versa.
        """
        data = super().to_representation(instance)
        self._round_decimals(data, instance)
        data["to"] = data.pop("to_address")
        data["from"] = data.pop("from_address")
        if data["kind"] == Transaction.KIND.deposit:
            data["deposit_memo_type"] = data["memo_type"]
            data["deposit_memo"] = data["memo"]
        else:
            data["withdraw_memo_type"] = data["memo_type"]
            data["withdraw_memo"] = data["memo"]
            data["withdraw_anchor_account"] = data["receiving_anchor_account"]
            del data["claimable_balance_id"]
        if (
            instance.protocol == Transaction.PROTOCOL.sep6
            and not settings.SEP6_USE_MORE_INFO_URL
        ):
            del data["more_info_url"]
        del data["memo_type"]
        del data["memo"]
        del data["receiving_anchor_account"]
        if instance.quote:
            data["amount_in_asset"] = instance.quote.sell_asset
            data["amount_out_asset"] = instance.quote.buy_asset
            if instance.fee_asset:
                data["amount_fee_asset"] = instance.fee_asset
        return data

    def _round_decimals(self, data, instance):
        """
        Rounds each decimal field to instance.asset.significant_decimals.
        """
        for field in ["amount_in", "amount_out", "amount_fee"]:
            if getattr(instance, field) is None:
                continue
            value = getattr(instance, field)
            if self.asset.significant_decimals == 7 and Decimal(
                "0.000001"
            ) > value >= Decimal("0.0000001"):
                # the decimal.Decimal class uses exponent notation for numbers
                # smaller than 0.000001 (6 decimals). Stellar only supports 7
                # decimals of precision, leaving 9 possible values where the
                # Decimal class uses exponent notation, which is corrected here.
                data[field] = f"{value:.7f}"
            else:
                data[field] = str(
                    round(getattr(instance, field), self.asset.significant_decimals)
                )

    class Meta:
        model = Transaction
        fields = [
            "id",
            "kind",
            "status",
            "status_eta",
            "amount_in",
            "amount_out",
            "amount_fee",
            "started_at",
            "completed_at",
            "stellar_transaction_id",
            "external_transaction_id",
            "from_address",
            "to_address",
            "receiving_anchor_account",
            "memo",
            "memo_type",
            "more_info_url",
            "refunded",
            "message",
            "claimable_balance_id",
        ]



import json
from decimal import Decimal, DecimalException

from django.urls import reverse
from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.utils.translation import gettext as _
from django.conf import settings as django_settings

from polaris.locale.utils import (
    activate_lang_for_request,
    validate_or_use_default_language,
)
from polaris import settings as polaris_settings
from polaris.templates import Template
from polaris.utils import render_error_response, getLogger
from polaris.models import Transaction, Asset, OffChainAsset
from polaris.integrations import registered_fee_func
from polaris.sep24.utils import verify_valid_asset_operation, get_timezone_utc_offset
from polaris.shared.serializers import TransactionSerializer
from polaris.integrations import (
    registered_deposit_integration as rdi,
    registered_withdrawal_integration as rwi,
)
from polaris.sep10.utils import SEP10Token


logger = getLogger(__name__)
SEP6_MORE_INFO_PATH = "/sep6/transaction/more_info"


def more_info(request: Request, sep6: bool = False) -> Response:
    try:
        transaction = _get_transaction_from_request(request, sep6=sep6)
    except (AttributeError, ValidationError) as exc:
        return render_error_response(str(exc), as_html=True)
    except ObjectDoesNotExist:
        return render_error_response(
            _("transaction not found"),
            status_code=status.HTTP_404_NOT_FOUND,
            as_html=True,
        )

    current_offset = get_timezone_utc_offset(
        request.session.get("timezone") or django_settings.TIME_ZONE
    )
    # persists the session, generating r.session.session_key
    #
    # this session key is passed to the rendered views and
    # used in client-side JavaScript in requests to the server
    if request.session.is_empty():
        request.session["authenticated"] = False
    else:
        request.session.modified = True
    if not request.session.session_key:
        request.session.create()
    serializer = TransactionSerializer(
        transaction, context={"request": request, "sep6": sep6}
    )
    tx_json = json.dumps({"transaction": serializer.data})
    context = {
        "tx_json": tx_json,
        "amount_in_asset": transaction.asset.asset_identification_format,
        "amount_out_asset": transaction.asset.asset_identification_format,
        "amount_in": serializer.data.get("amount_in"),
        "amount_out": serializer.data.get("amount_out"),
        "amount_fee": serializer.data.get("amount_fee"),
        "amount_in_symbol": transaction.asset.symbol,
        "amount_fee_symbol": transaction.asset.symbol,
        "amount_out_symbol": transaction.asset.symbol,
        "amount_in_significant_decimals": transaction.asset.significant_decimals,
        "amount_fee_significant_decimals": transaction.asset.significant_decimals,
        "amount_out_significant_decimals": transaction.asset.significant_decimals,
        "transaction": transaction,
        "asset": transaction.asset,
        "offchain_asset": None,
        "price": None,
        "price_inversion": None,
        "price_inversion_significant_decimals": None,
        "exchange_amount": None,
        "exchanged_amount": None,
        "current_offset": current_offset,
        "timezone_endpoint": reverse("tzinfo"),
        "session_id": request.session.session_key,
    }
    if transaction.quote:
        if "deposit" in transaction.kind:
            scheme, identifier = transaction.quote.sell_asset.split(":")
        else:
            scheme, identifier = transaction.quote.buy_asset.split(":")
        offchain_asset = OffChainAsset.objects.get(scheme=scheme, identifier=identifier)
        if "deposit" in transaction.kind:
            context.update(
                **{
                    "amount_in_asset": offchain_asset.asset_identification_format,
                    "amount_in_symbol": offchain_asset.symbol,
                    "amount_in_significant_decimals": offchain_asset.significant_decimals,
                }
            )
        else:
            context.update(
                **{
                    "amount_out_asset": offchain_asset.asset_identification_format,
                    "amount_out_symbol": offchain_asset.symbol,
                    "amount_out_significant_decimals": offchain_asset.significant_decimals,
                }
            )
        if transaction.fee_asset == offchain_asset.asset_identification_format:
            context.update(
                **{
                    "amount_fee_symbol": offchain_asset.symbol,
                    "amount_fee_significant_decimals": offchain_asset.significant_decimals,
                }
            )
        price_inversion = 1 / transaction.quote.price
        price_inversion_sd = min(
            transaction.asset.significant_decimals, offchain_asset.significant_decimals
        )
        while (
            calc_amount_out_with_price_inversion(
                transaction, price_inversion, price_inversion_sd, context
            )
            != transaction.amount_out
            and price_inversion_sd < 7
        ):
            price_inversion_sd += 1
        if (
            transaction.fee_asset == offchain_asset.asset_identification_format
            and "deposit" in transaction.kind
        ) or (
            transaction.fee_asset == transaction.asset.asset_identification_format
            and "withdrawal" in transaction.kind
        ):
            context["exchange_amount"] = transaction.amount_in - transaction.amount_fee
        else:
            context["exchanged_amount"] = round(
                transaction.amount_in * price_inversion,
                context["amount_out_significant_decimals"],
            )
        context.update(
            **{
                "offchain_asset": offchain_asset,
                "price": transaction.quote.price,
                "price_inversion": round(
                    1 / transaction.quote.price, price_inversion_sd
                ),
                "price_inversion_significant_decimals": price_inversion_sd,
            }
        )

    integration_class = rdi if "deposit" in transaction.kind else rwi
    try:
        content_from_anchor = (
            integration_class.content_for_template(
                request=request,
                template=Template.MORE_INFO,
                transaction=transaction,
            )
            or {}
        )
    except NotImplementedError:
        content_from_anchor = {}

    context.update(content_from_anchor)

    # more_info.html will update the 'callback' parameter value to 'success' after
    # making the callback. If the page is then reloaded, the callback is not included
    # in the rendering context, ensuring only one successful callback request is made.
    callback = request.GET.get("callback")
    if callback and callback != "success":
        context["callback"] = callback

    return Response(
        context,
        template_name=content_from_anchor.get(
            "template_name", "polaris/more_info.html"
        ),
    )


def calc_amount_out_with_price_inversion(
    transaction: Transaction,
    price_inversion: Decimal,
    price_inversion_significant_decimals: int,
    context: dict,
):
    if (
        transaction.fee_asset == transaction.asset.asset_identification_format
        and "deposit" not in transaction.kind
    ) or (
        transaction.fee_asset != transaction.asset.asset_identification_format
        and "deposit" in transaction.kind
    ):
        return round(
            round(price_inversion, price_inversion_significant_decimals)
            * (transaction.amount_in - transaction.amount_fee),
            context["amount_out_significant_decimals"],
        )
    else:
        return round(
            round(price_inversion, price_inversion_significant_decimals)
            * transaction.amount_in
            - transaction.amount_fee,
            context["amount_out_significant_decimals"],
        )


def transactions_request(
    request: Request,
    token: SEP10Token,
    sep6: bool = False,
) -> Response:
    activate_lang_for_request(validate_or_use_default_language(request.GET.get("lang")))

    try:
        limit = _validate_limit(request.GET.get("limit"))
    except ValueError:
        return render_error_response(
            "invalid limit", status_code=status.HTTP_400_BAD_REQUEST
        )

    protocol_filter = {"sep6_enabled": True} if sep6 else {"sep24_enabled": True}
    if not request.GET.get("asset_code"):
        return render_error_response("asset_code is required")
    elif not Asset.objects.filter(
        code=request.GET.get("asset_code"), **protocol_filter
    ).exists():
        return render_error_response("invalid asset_code")

    translation_dict = {
        "asset_code": "asset__code",
        "no_older_than": "started_at__gte",
        "kind": "kind",
    }

    qset_filter = _compute_qset_filters(request.GET, translation_dict)
    qset_filter["stellar_account"] = token.account
    qset_filter["muxed_account"] = token.muxed_account
    qset_filter["account_memo"] = token.memo

    # Since the Transaction IDs are UUIDs, rather than in the chronological
    # order of their creation, we map the paging ID (if provided) to the
    # started_at field of a Transaction.
    paging_id = request.GET.get("paging_id")
    if paging_id:
        try:
            start_transaction = Transaction.objects.get(id=paging_id)
        except ObjectDoesNotExist:
            return render_error_response(
                "invalid paging_id", status_code=status.HTTP_400_BAD_REQUEST
            )
        qset_filter["started_at__lt"] = start_transaction.started_at

    protocol = Transaction.PROTOCOL.sep6 if sep6 else Transaction.PROTOCOL.sep24
    transactions_qset = Transaction.objects.filter(protocol=protocol, **qset_filter)
    if limit:
        transactions_qset = transactions_qset[:limit]

    serializer = TransactionSerializer(
        transactions_qset,
        many=True,
        context={"request": request, "same_asset": True, "sep6": sep6},
    )

    return Response({"transactions": serializer.data})


def transaction_request(
    request: Request,
    token: SEP10Token,
    sep6: bool = False,
) -> Response:
    activate_lang_for_request(validate_or_use_default_language(request.GET.get("lang")))
    try:
        request_transaction = _get_transaction_from_request(
            request,
            token=token,
            sep6=sep6,
        )
    except (AttributeError, ValidationError) as exc:
        return render_error_response(str(exc), status_code=status.HTTP_400_BAD_REQUEST)
    except ObjectDoesNotExist:
        return render_error_response(
            "transaction not found", status_code=status.HTTP_404_NOT_FOUND
        )
    serializer = TransactionSerializer(
        request_transaction,
        context={"request": request, "sep6": sep6},
    )
    return Response({"transaction": serializer.data})


def fee(request: Request, sep6: bool = False) -> Response:
    """
    Definition of the /fee endpoint, in accordance with SEP-0024.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#fee
    """
    deposit_op = polaris_settings.OPERATION_DEPOSIT
    withdrawal_op = polaris_settings.OPERATION_WITHDRAWAL

    operation = request.GET.get("operation")
    op_type = request.GET.get("type")
    asset_code = request.GET.get("asset_code")
    amount_str = request.GET.get("amount")

    # Verify that the asset code exists in our database:
    protocol_filter = {"sep6_enabled": True} if sep6 else {"sep24_enabled": True}
    asset = Asset.objects.filter(code=asset_code, **protocol_filter).first()
    if not asset_code or not asset:
        return render_error_response("invalid 'asset_code'")

    # Verify that amount is provided, and that can be parsed into a decimal:
    try:
        amount = Decimal(amount_str)
    except (DecimalException, TypeError):
        return render_error_response("invalid 'amount'")

    error_resp = None
    # Verify that the requested operation is valid:
    if operation not in (deposit_op, withdrawal_op):
        error_resp = render_error_response(
            f"'operation' should be either '{deposit_op}' or '{withdrawal_op}'"
        )
    # Verify asset is enabled and within the specified limits
    elif operation == deposit_op:
        error_resp = verify_valid_asset_operation(
            asset, amount, Transaction.KIND.deposit
        )
    elif operation == withdrawal_op:
        error_resp = verify_valid_asset_operation(
            asset, amount, Transaction.KIND.withdrawal
        )

    if error_resp:
        return error_resp

    try:
        fee_amount = registered_fee_func(
            request=request,
            fee_params={
                "operation": operation,
                "type": op_type,
                "asset_code": asset_code,
                "amount": amount,
            },
        )
    except ValueError:
        return render_error_response(
            _("unable to calculate fees for the requested asset")
        )

    return Response({"fee": fee_amount})


def _validate_limit(limit):
    if not limit:
        return limit
    limit = int(limit)
    if limit < 1:
        raise ValueError
    return limit


def _compute_qset_filters(req_params, translation_dict):
    """
    _compute_qset_filters translates the keys of req_params to the keys of translation_dict.
    If the key isn't present in filters_dict, it is discarded.
    """
    return {
        translation_dict[rp]: req_params[rp]
        for rp in filter(lambda i: i in translation_dict, req_params.keys())
    }


def _get_transaction_from_request(
    request,
    token: SEP10Token = None,
    sep6: bool = False,
):
    translation_dict = {
        "id": "id",
        "stellar_transaction_id": "stellar_transaction_id",
        "external_transaction_id": "external_transaction_id",
    }

    qset_filter = _compute_qset_filters(request.GET, translation_dict)
    if not qset_filter:
        raise AttributeError(
            _(
                "at least one of id, stellar_transaction_id, or "
                "external_transaction_id must be provided"
            )
        )

    if token:
        qset_filter["stellar_account"] = token.account
        qset_filter["muxed_account"] = token.muxed_account
        qset_filter["account_memo"] = token.memo

    protocol = Transaction.PROTOCOL.sep6 if sep6 else Transaction.PROTOCOL.sep24
    return Transaction.objects.get(protocol=protocol, **qset_filter)



"""This module defines the logic for the `/info` endpoint."""
from typing import Dict

from rest_framework.decorators import api_view, renderer_classes
from rest_framework.response import Response
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer

from polaris.locale.utils import (
    activate_lang_for_request,
    validate_or_use_default_language,
)
from polaris.models import Asset
from polaris.integrations import (
    registered_custody_integration as rci,
)


def _get_asset_info(asset: Asset, op_type: str) -> Dict:
    if not getattr(asset, f"{op_type}_enabled"):
        return {"enabled": False}

    asset_info = {"enabled": True}
    min_amount_attr = f"{op_type}_min_amount"
    max_amount_attr = f"{op_type}_max_amount"
    min_amount = getattr(asset, min_amount_attr)
    max_amount = getattr(asset, max_amount_attr)
    if min_amount > getattr(Asset, "_meta").get_field(min_amount_attr).default:
        asset_info["min_amount"] = min_amount
    if max_amount < getattr(Asset, "_meta").get_field(max_amount_attr).default:
        asset_info["max_amount"] = max_amount
    if getattr(asset, f"{op_type}_fee_fixed") is not None:
        asset_info["fee_fixed"] = getattr(asset, f"{op_type}_fee_fixed")
    if getattr(asset, f"{op_type}_fee_percent") is not None:
        asset_info["fee_percent"] = getattr(asset, f"{op_type}_fee_percent")

    return asset_info


@api_view()
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
def info(request):
    """
    Definition of the /info endpoint, in accordance with SEP-0024.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#info
    """
    activate_lang_for_request(validate_or_use_default_language(request.GET.get("lang")))
    info_data = {
        "deposit": {},
        "withdraw": {},
        "fee": {"enabled": True},
        "features": {
            "account_creation": rci.account_creation_supported,
            "claimable_balances": rci.claimable_balances_supported,
        },
    }

    for asset in Asset.objects.filter(sep24_enabled=True):
        info_data["deposit"][asset.code] = _get_asset_info(asset, "deposit")
        info_data["withdraw"][asset.code] = _get_asset_info(asset, "withdrawal")

    return Response(info_data)



from corsheaders.signals import check_request_enabled


def cors_allow_origins_for_polaris_requests(sender, request, **_kwargs):
    return (
        request.path.startswith("/sep24")
        or request.path.startswith("/sep6")
        or request.path.startswith("/sep31")
        or request.path.startswith("/sep38")
        or request.path.startswith("/.well-known")
        or request.path.startswith("/auth")
        or request.path.startswith("/kyc")
    )


check_request_enabled.connect(cors_allow_origins_for_polaris_requests)



import sys
import signal
import time
from datetime import datetime, timezone

from django.core.management import BaseCommand

from polaris.utils import getLogger, maybe_make_callback
from polaris.models import Transaction
from polaris.integrations import registered_rails_integration as rri


logger = getLogger(__name__)
DEFAULT_INTERVAL = 30
TERMINATE = False


class Command(BaseCommand):
    """
    Polaris periodically queries for transactions in ``pending_external`` and passes
    them to the :meth:`~polaris.integrations.RailsIntegration.poll_outgoing_transactions`.

    The anchor is expected to update the transactions’
    :attr:`~polaris.models.Transaction.status` to ``completed`` if the funds have been
    confirmed to be delivered.

    **Optional arguments:**

        -h, --help            show this help message and exit
        --loop                Continually restart command after a specified number
                              of seconds.
        --interval INTERVAL, -i INTERVAL
                              The number of seconds to wait before restarting
                              command. Defaults to 30.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    @staticmethod
    def exit_gracefully(sig, frame):  # pragma: no cover
        logger.info("Exiting poll_outgoing_transactions...")
        module = sys.modules[__name__]
        module.TERMINATE = True

    @staticmethod
    def sleep(seconds):  # pragma: no cover
        module = sys.modules[__name__]
        for _ in range(seconds):
            if module.TERMINATE:
                break
            time.sleep(1)

    def add_arguments(self, parser):  # pragma: no cover
        parser.add_argument(
            "--loop",
            action="store_true",
            help="Continually restart command after a specified number of seconds.",
        )
        parser.add_argument(
            "--interval",
            "-i",
            type=int,
            help=(
                "The number of seconds to wait before restarting command. "
                "Defaults to {}.".format(DEFAULT_INTERVAL)
            ),
        )

    def handle(self, *_args, **options):  # pragma: no cover
        module = sys.modules[__name__]
        if options.get("loop"):
            while True:
                if module.TERMINATE:
                    break
                self.poll_outgoing_transactions()
                self.sleep(options.get("interval") or DEFAULT_INTERVAL)
        else:
            self.poll_outgoing_transactions()

    @staticmethod
    def poll_outgoing_transactions():
        transactions = Transaction.objects.filter(
            kind__in=[
                Transaction.KIND.withdrawal,
                getattr(Transaction.KIND, "withdrawal-exchange"),
                Transaction.KIND.send,
            ],
            status=Transaction.STATUS.pending_external,
        )
        try:
            complete_transactions = rri.poll_outgoing_transactions(transactions)
        except NotImplementedError:
            module = sys.modules[__name__]
            module.TERMINATE = True
            return
        except Exception:
            logger.exception("An exception was raised by poll_pending_transfers()")
            return

        if not (
            isinstance(complete_transactions, list)
            and all(isinstance(t, Transaction) for t in complete_transactions)
        ):
            logger.exception(
                "invalid return type, expected a list of Transaction objects"
            )
            return

        ids = [t.id for t in complete_transactions]
        if ids:
            num_completed = Transaction.objects.filter(id__in=ids).update(
                status=Transaction.STATUS.completed,
                completed_at=datetime.now(timezone.utc),
            )
            logger.info(f"{num_completed} pending transfers have been completed")
        for t in complete_transactions:
            maybe_make_callback(t)



from typing import Optional, List, Dict

from rest_framework.request import Request

from polaris.sep10.token import SEP10Token


class CustomerIntegration:
    def more_info_url(
        self,
        token: SEP10Token,
        request: Request,
        account: str,
        *args: List,
        memo: Optional[int] = None,
        **kwargs: Dict
    ) -> str:
        """
        .. _SEP-6 Customer Information Status: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#4-customer-information-status

        Return a URL the client can open in a browser to view the status of their account
        with the anchor. This URL will be returned in a `SEP-6 Customer Information Status`_
        response. This is optional.

        :param token: the ``SEP10Token`` object representing the authenticated session
        :param request: a ``rest_framework.request.Request`` instance
        :param account: the stellar or muxed account for the url to be returned
        :param memo: the ID memo used to identify the user of the shared Stellar account
        """
        raise NotImplementedError()

    def get(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        *args: List,
        **kwargs: Dict
    ) -> Dict:
        """
        .. _`SEP-12 GET /customer`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0012.md#customer-get
        .. _ObjectDoesNotExist: https://docs.djangoproject.com/en/3.1/ref/exceptions/#objectdoesnotexist

        Return a dictionary matching the response schema outlined in `SEP-12 GET /customer`_
        based on the `params` passed.

        Raise a ``ValueError`` if the parameters are invalid, or raise an
        ObjectDoesNotExist_ exception if the customer specified via the ``id`` parameter
        does not exist. An error response with the appropriate status will be sent using
        the message passed to the exception.

        :param token: the ``SEP10Token`` object representing the authenticated session
        :param request: a ``rest_framework.request.Request`` instance
        :param params: request parameters as described in SEP-12
        """
        raise NotImplementedError()

    def put(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        *args: List,
        **kwargs: Dict
    ) -> str:
        """
        Update or create a record of the customer information passed. This information can
        then later be queried for when a client requests a deposit or withdraw on behalf of
        the customer.

        If the information passed in `params` is invalid in some way, raise a ``ValueError``
        for Polaris to return a 400 Bad Request response to the client. The message contained
        in the exception will be passed as the error message in the response. If the request
        specified a customer `id` in the request body but a record with that ID doesn't exist,
        raise a ``django.exceptions.ObjectDoesNotExist`` exception. Polaris will return a 404
        response.

        Return a customer ID that clients can use in future requests, such as a `GET /customer`
        request or a SEP-31 `POST /transactions` request. If the request included an `id`
        parameter, make sure the same id is returned.

        If the required information is provided and the customer has ``Transaction`` objects
        in the ``pending_customer_info_update`` status, all such ``Transaction.status`` values
        should be updated to ``pending_receiver``. Polaris will then call the
        ``execute_outgoing_transaction`` integration function for each updated transaction.

        If the anchor requires verification of values passed in `params`, such as
        ``mobile_number`` or ``email_address``, the anchor should initate the verification
        process in this function and ensure the next call to ``GET /customer`` contains the
        field requiring verification with a ``VERIFICATION_REQUIRED`` status. Unless the anchor
        expects users to verify the field value in some other manner, such as opening a
        confirmation link sent via email, client applications will eventually make a call to
        ``PUT /customer/verification`` to return the verification codes sent as a result of the
        initial call to ``PUT /customer``.

        :param token: the ``SEP10Token`` object representing the authenticated session
        :param request: a ``rest_framework.request.Request`` instance
        :param params: request parameters as described in SEP-12
        :raises: ValueError or ObjectDoesNotExist
        """
        raise NotImplementedError()

    def delete(
        self,
        token: SEP10Token,
        request: Request,
        account: str,
        memo: Optional[str],
        memo_type: Optional[str],
        *args: List,
        **kwargs: Dict
    ):
        """
        Delete the record of the customer specified by `account`, `memo`, and `memo_type`.
        If such a record does not exist, raise a ``ObjectDoesNotExist`` exception for Polaris
        to return a 404 Not Found response.

        :param token: the ``SEP10Token`` object representing the authenticated session
        :param request: a ``rest_framework.request.Request`` instance
        :param account: the stellar or muxed account associated with the customer
        :param memo: the optional memo used to create the customer
        :param memo_type: the optional type of the memo used to create to the customer
        """
        raise NotImplementedError()

    def callback(
        self,
        token: SEP10Token,
        request: Request,
        params: Dict,
        *args: List,
        **kwargs: Dict
    ):
        """
        Save the URL provided in association with the user identified by the parameters sent
        in the request. The anchor is responsible for making POST requests containing the
        response body of a GET request to the saved URL whenever the SEP-12 status of the
        customer changes. Polaris does not manage an anchor's customer data and therefore
        cannot make these requests.

        Client applications may register callback URLs if the application does not have the
        ability to poll the ``GET /customer`` endopoint at any time, which requires SEP-10
        authentication (and consequently the relevant account's signature).

        If the customer specified does not exist, raise an ``ObjectDoesNotExist``. If the
        URL is provided is invalid in some way, raise a ``ValueError``.

        If this function is not implemented, Polaris will respond with a 501 Not Implemented.

        :param token: the ``SEP10Token`` object representing the authenticated session
        :param request: a ``rest_framework.request.Request`` instance
        :param params: request parameters as described in SEP-12
        :raises: ValueError or django.core.exceptions.ObjectDoesNotExist
        """
        raise NotImplementedError()

    def put_verification(
        self,
        token: SEP10Token,
        request: Request,
        account: str,
        params: Dict,
        *args: List,
        **kwargs: Dict
    ) -> Dict:
        """
        .. _`endpoint specification`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0012.md#customer-put-verification

        Validate the values, typically verification codes, passed in `params` for the
        customer identified by ``params["id"]``. See the `endpoint specification`_ for
        more information on the request format.

        Anchors may return fields from ``GET /customer`` requests with the
        ``VERIFICATION_REQUIRED`` status if the anchor requires the user to verify a
        SEP-9 field value provided in a previous call to ``PUT /customer``. The most
        common field needing verification is `mobile_number`. This function will be
        called when the client passes the verification value back to the anchor.

        If the validation values are correct, return a dictionary that is identical to
        what would be returned for a call to ``GET /customer``.

        If any of the validation values are incorrect, raise a ``ValueError`` and
        Polaris will raise a 400 Bad Request.

        If the customer specified by ``params["id"]`` does not exist, or the authenticated
        ``account`` does not match Stellar account associated with the ID, raise an
        ``ObjectDoesNotExist`` exception. Polaris will return a 404 Not Found response.

        If this function is not implemented, Polaris will respond with a 501 Not Implemented.

        :param token: the ``SEP10Token`` object representing the authenticated session
        :param request: a ``rest_framework.request.Request`` instance
        :param account: the Stellar or muxed account authenticated via SEP-10
        :param params: the request body of the ``PUT /customer/verification`` call
        :raises: ValueError, ObjectDoesNotExist, NotImplementedError
        """
        raise NotImplementedError()


registered_customer_integration = CustomerIntegration()



from decimal import Decimal
from typing import List, Union, Optional

from rest_framework.request import Request

from polaris.sep10.token import SEP10Token
from polaris.models import Quote, OffChainAsset, Asset, DeliveryMethod


class QuoteIntegration:
    def get_prices(
        self,
        token: SEP10Token,
        request: Request,
        sell_asset: Union[Asset, OffChainAsset],
        sell_amount: Decimal,
        buy_assets: List[Union[Asset, OffChainAsset]],
        sell_delivery_method: Optional[DeliveryMethod] = None,
        buy_delivery_method: Optional[DeliveryMethod] = None,
        country_code: Optional[str] = None,
        *args,
        **kwargs
    ) -> List[Decimal]:
        """
        Return a list of prices in the order of the `buy_assets` provided. Each price
        should be the value of one unit of the `buy_asset` in terms of `sell_asset`.
        The prices returned from this function are non-binding, meaning a ``Quote``
        object will not be created and returned to the client as a result of this call.

        The `buy_assets` list passed are the assets that have an ``ExchangePair``
        database record with the `sell_asset`. Polaris will also ensure all `buy_assets`
        support the `buy_delivery_method` and `country_code` if specified by the client.

        Polaris will also ensure that the `sell_delivery_method` is supported for
        `sell_asset`.

        Raise a ``ValueError`` if the parameters are invalid in some way. Because Polaris
        validates all the parameters passed, the only reason to raise this exception
        _should_ be because the `sell_amount` is outside the minimum and maximum bounds
        for your service. Polaris will return a 400 Bad Request status code in this case.

        Raise a ``RuntimeError`` if you cannot return prices for the provided `sell_asset`
        and `buy_assets` for any reason. For example, the service used by the anchor to
        source exchange prices could be down. Polaris will return a 503 Server Unavailable
        status code in this case.

        :param token: The ``SEP10Token`` object representing the authenticated session
        :param request: The ``rest_framework.Request`` object representing the request
        :param sell_asset: The asset the client would like to sell for `buy_assets`
        :param buy_assets: The assets the client would like to buy using `sell_asset`
        :param sell_amount: The amount the client would like to sell of `sell_asset`
        :param sell_delivery_method: The method the client would like to use to deliver
            funds to the anchor.
        :param buy_delivery_method: The method the client would like to use to receive
            or collect funds from the anchor.
        :param country_code: The ISO 3166-1 alpha-3 code of the user's current address
        """
        raise NotImplementedError()

    def get_price(
        self,
        token: SEP10Token,
        request: Request,
        sell_asset: Union[Asset, OffChainAsset],
        buy_asset: Union[Asset, OffChainAsset],
        buy_amount: Optional[Decimal] = None,
        sell_amount: Optional[Decimal] = None,
        sell_delivery_method: Optional[DeliveryMethod] = None,
        buy_delivery_method: Optional[DeliveryMethod] = None,
        country_code: Optional[str] = None,
        *args,
        **kwargs
    ) -> Decimal:
        """
        Return the price of one unit of `buy_asset` in terms of `sell_asset`. The price
        returned from this function is non-binding, meaning a ``Quote`` object will not
        be created and returned to the client as a result of this call.

        Polaris will ensure that there is an ``ExchangePair`` database record for the
        `sell_asset` and `buy_asset`, that the specified `sell_delivery_method` or
        `buy_delivery_method` is supported by the off-chain asset, and that the anchor
        supports transacting the off-chain asset in the `country_code`, if specified.

        Raise a ``ValueError`` if the parameters are invalid in some way. Because Polaris
        validates all the parameters passed, the only reason to raise this exception
        _should_ be because `sell_amount` or `buy_amount` is outside the minimum and
        maximum bounds for your service. Polaris will return a 400 Bad Request status
        code in this case.

        Raise a ``RuntimeError`` if you cannot return prices for the provided `sell_asset`
        and `buy_asset` for any reason. For example, the service used by the anchor to
        source exchange prices could be down. Polaris will return a 503 Server Unavailable
        status code in this case.

        :param token: The ``SEP10Token`` object representing the authenticated session
        :param request: The ``rest_framework.Request`` object representing the request
        :param sell_asset: The asset the client would like to sell for `buy_assets`
        :param sell_amount: The amount the client would like to sell of `sell_asset`
        :param buy_asset: The asset the client would like to buy using `sell_asset`
        :param buy_amount: The amount the client would like to purchase of `buy_asset`
        :param sell_delivery_method: The method the client would like to use to deliver
            funds to the anchor.
        :param buy_delivery_method: The method the client would like to use to receive
            or collect funds from the anchor.
        :param country_code: The ISO 3166-1 alpha-3 code of the user's current address
        """
        raise NotImplementedError()

    def post_quote(
        self, token: SEP10Token, request: Request, quote: Quote, *args, **kwargs
    ) -> Quote:
        """
        Assign ``Quote.price`` and Quote.expire_at`` on the `quote` passed and return it.
        The anchor will be expected to honor the price set when the quote is used in a
        ``Transaction``. Note that the ``Quote`` object passed is not yet saved to the
        database when this function is called. If no exception is raised, Polaris will
        calculate the amount of the asset not specified in the request using the price
        assigned and save the returned quote to the database. However, the anchor is free
        to save the quote to the database in this function if necessary.

        Polaris will ensure that there is an ``ExchangePair`` database record for the
        `sell_asset` and `buy_asset`, that the specified `sell_delivery_method` or
        `buy_delivery_method` is supported by the off-chain asset, and that the anchor
        supports transacting the off-chain asset in the `country_code`, if specified.

        Raise a ``ValueError`` if the parameters are invalid in some way. Because Polaris
        validates all the parameters passed, the only reason to raise this exception
        _should_ be because `sell_amount` or `buy_amount` is outside the minimum and
        maximum bounds for your service OR the requested `expires_at` value is not
        acceptable to the anchor. Polaris will return a 400 Bad Request status code in
        this case.

        Raise a ``RuntimeError`` if you cannot return prices for the provided `sell_asset`
        and `buy_asset` for any reason. For example, the service used by the anchor to
        source exchange prices could be down. Polaris will return a 503 Server Unavailable
        status code in this case.

        :param token: The ``SEP10Token`` object representing the authenticated session
        :param request: The ``rest_framework.Request`` object representing the request
        :param quote: The ``Quote`` object representing the exchange rate to be offered
        """
        raise NotImplementedError()


registered_quote_integration = QuoteIntegration()



# Generated by Django 2.2.24 on 2021-08-09 16:05

from django.db import migrations, models
import polaris.models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0009_transaction_client_domain"),
    ]

    operations = [
        migrations.AlterField(
            model_name="asset",
            name="distribution_seed",
            field=polaris.models.EncryptedTextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="completed_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="paging_token",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="protocol",
            field=models.CharField(
                blank=True,
                choices=[("sep6", "sep6"), ("sep24", "sep24"), ("sep31", "sep31")],
                max_length=5,
                null=True,
            ),
        ),
    ]



from typing import Dict
from polaris.utils import getLogger

from django.utils.translation import gettext as _
from rest_framework.decorators import api_view, renderer_classes
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer

from polaris.models import Asset
from polaris.utils import render_error_response
from polaris.integrations import registered_sep31_receiver_integration
from polaris import settings


logger = getLogger(__name__)


@api_view(["GET"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
def info(request: Request) -> Response:
    info_data = {
        "receive": {},
    }
    for asset in Asset.objects.filter(sep31_enabled=True):
        try:
            fields_and_types = registered_sep31_receiver_integration.info(
                request=request, asset=asset, lang=request.GET.get("lang")
            )
        except ValueError:
            return render_error_response("unsupported 'lang'")
        try:
            validate_info_response(fields_and_types)
        except ValueError as e:
            logger.error(f"info integration error: {str(e)}")
            return render_error_response(
                _("unable to process the request"),
                status_code=500,
            )
        info_data["receive"][asset.code] = get_asset_info(asset, fields_and_types)

    return Response(info_data)


def validate_info_response(fields_and_types: Dict):
    if not isinstance(fields_and_types, dict):
        raise ValueError("info integration must return a dictionary")
    elif not all(
        f in ["fields", "sep12", "sender_sep12_type", "receiver_sep12_type"]
        for f in fields_and_types.keys()
    ):
        raise ValueError("unrecognized key in info integration response")
    elif "fields" not in fields_and_types:
        raise ValueError("missing fields object in info response")
    elif not isinstance(fields_and_types["fields"], dict):
        raise ValueError("invalid type for fields value")
    validate_fields(fields_and_types["fields"].get("transaction"))

    if (
        "sender_sep12_type" in fields_and_types
        or "receiver_sep12_type" in fields_and_types
    ):
        if "sender_sep12_type" in fields_and_types and not isinstance(
            fields_and_types["sender_sep12_type"], str
        ):
            raise ValueError("sender_sep12_type must be a string")
        if "receiver_sep12_type" in fields_and_types and not isinstance(
            fields_and_types["receiver_sep12_type"], str
        ):
            raise ValueError("receiver_sep12_type must be a string")
        if "sep12" in fields_and_types:
            raise ValueError(
                "cannot specify both sep12 object and sender_sep12_type/receiver_sep12_type"
            )
        return

    if "sep12" not in fields_and_types:
        raise ValueError("missing sep12 object in info response")
    elif not isinstance(fields_and_types["sep12"], dict):
        raise ValueError("invalid type for sep12 key value")
    elif not (
        "sender" in fields_and_types["sep12"]
        and "receiver" in fields_and_types["sep12"]
    ):
        raise ValueError("sender and/or receiver object missing in sep12 object")
    elif not (
        isinstance(fields_and_types["sep12"]["sender"], dict)
        and isinstance(fields_and_types["sep12"]["receiver"], dict)
    ):
        raise ValueError("sender and receiver key values must be objects")
    validate_types(fields_and_types["sep12"]["sender"].get("types"))
    validate_types(fields_and_types["sep12"]["receiver"].get("types"))


def validate_fields(field_dict: Dict):
    if not field_dict:
        return
    elif not isinstance(field_dict, dict):
        raise ValueError("fields key value must be a dict")
    for key, val in field_dict.items():
        if not isinstance(val, dict):
            raise ValueError(f"{key} value must be a dict, got {type(val)}")
        desc = val.get("description")
        optional = val.get("optional")
        choices = val.get("choices")
        if not desc:
            raise ValueError("'fields' dict must contain 'description'")
        if not set(val.keys()).issubset({"description", "optional", "choices"}):
            raise ValueError("unexpected keys in 'fields' dict")
        if not isinstance(desc, str):
            raise ValueError("'description' must be a string")
        if optional and not isinstance(optional, bool):
            raise ValueError("'optional' must be a boolean")
        if choices and not isinstance(choices, list):
            raise ValueError("'choices' must be a list")


def validate_types(types: Dict):
    if not types:
        return
    elif not isinstance(types, dict):
        raise ValueError("types key value must be an dict")
    for key, val in types.items():
        if not isinstance(val, dict):
            raise ValueError(f"{key} value must be a dict, got {type(val)}")
        if "description" not in val:
            raise ValueError(f"{key} dict must contain a description")
        if not isinstance(val["description"], str):
            raise ValueError(f"{key} description must be a human-readable string")
        if len(val) != 1:
            raise ValueError(f"unexpected key in {key} dict")


def get_asset_info(asset: Asset, fields_and_types: Dict) -> Dict:
    if not asset.sep31_enabled:
        return {"enabled": False}

    asset_info = {
        "enabled": True,
        **fields_and_types,
    }
    min_amount = getattr(asset, "send_min_amount")
    max_amount = getattr(asset, "send_max_amount")
    if min_amount > getattr(Asset, "_meta").get_field("send_min_amount").default:
        asset_info["min_amount"] = min_amount
    if max_amount < getattr(Asset, "_meta").get_field("send_max_amount").default:
        asset_info["max_amount"] = max_amount
    if asset.send_fee_fixed is not None:
        asset_info["fee_fixed"] = round(
            asset.send_fee_fixed, asset.significant_decimals
        )
    if asset.send_fee_percent is not None:
        asset_info["fee_percent"] = asset.send_fee_percent
    if "sep-38" in settings.ACTIVE_SEPS:
        asset_info["quotes_supported"] = asset.sep38_enabled

    return asset_info



"""This module defines the `/fee` view."""

from rest_framework.decorators import api_view, renderer_classes
from rest_framework.response import Response
from rest_framework.renderers import JSONRenderer
from rest_framework.request import Request

from polaris.shared import endpoints


@api_view(["GET"])
@renderer_classes([JSONRenderer])
def fee(request: Request) -> Response:
    """
    SEP-24 and SEP-6 /fee endpoints are identical
    """
    return endpoints.fee(request)



from typing import Dict, Optional, Tuple

from django.utils.translation import gettext as _
from django.core.validators import URLValidator
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from rest_framework.views import APIView
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.decorators import api_view, renderer_classes, parser_classes
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer
from rest_framework.parsers import JSONParser, MultiPartParser, FormParser

from polaris import settings
from polaris.models import Transaction
from polaris.utils import (
    extract_sep9_fields,
    render_error_response,
    make_memo,
    getLogger,
    SEP_9_FIELDS,
)
from polaris.sep10.utils import validate_sep10_token
from polaris.sep10.token import SEP10Token
from polaris.integrations import registered_customer_integration as rci


logger = getLogger(__name__)


class CustomerAPIView(APIView):
    renderer_classes = [JSONRenderer, BrowsableAPIRenderer]
    parser_classes = [MultiPartParser, FormParser, JSONParser]

    @staticmethod
    @validate_sep10_token()
    def get(token: SEP10Token, request: Request) -> Response:
        maybe_response = validate_id_parameters(
            token=token, request_data=request.query_params
        )
        if maybe_response:
            return maybe_response

        try:
            memo, memo_type = validate_memo_parameters(
                token=token, request_data=request.query_params
            )
        except ValueError as e:
            return render_error_response(str(e))

        try:
            response_data = rci.get(
                token=token,
                request=request,
                params={
                    "id": request.GET.get("id"),
                    "account": token.muxed_account or token.account,
                    "memo": memo,
                    "memo_type": memo_type,
                    "type": request.GET.get("type"),
                    "lang": request.GET.get("lang"),
                },
            )
        except ValueError as e:
            return render_error_response(str(e), status_code=400)
        except ObjectDoesNotExist as e:
            return render_error_response(str(e), status_code=404)

        try:
            validate_response_data(response_data)
        except ValueError:
            logger.exception(
                "An exception was raised validating GET /customer response"
            )
            return render_error_response(
                _("unable to process request."), status_code=500
            )

        return Response(response_data)

    @staticmethod
    @validate_sep10_token()
    def put(token: SEP10Token, request: Request) -> Response:
        maybe_response = validate_id_parameters(token=token, request_data=request.data)
        if maybe_response:
            return maybe_response

        try:
            memo, memo_type = validate_memo_parameters(
                token=token, request_data=request.data
            )
        except ValueError as e:
            return render_error_response(str(e))

        try:
            customer_id = rci.put(
                token=token,
                request=request,
                params={
                    "id": request.data.get("id"),
                    "account": token.muxed_account or token.account,
                    "memo": memo,
                    "memo_type": memo_type,
                    "type": request.data.get("type"),
                    **extract_sep9_fields(request.data),
                },
            )
        except ValueError as e:
            return render_error_response(str(e), status_code=400)
        except ObjectDoesNotExist as e:
            return render_error_response(str(e), status_code=404)

        if not isinstance(customer_id, str):
            logger.error(
                "Invalid customer ID returned from put() integration. Must be str."
            )
            return render_error_response(_("unable to process request"))

        return Response({"id": customer_id}, status=202)


@api_view(["PUT"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@parser_classes([MultiPartParser, FormParser, JSONParser])
@validate_sep10_token()
def callback(token: SEP10Token, request: Request) -> Response:
    maybe_response = validate_id_parameters(token=token, request_data=request.data)
    if maybe_response:
        return maybe_response

    try:
        memo, memo_type = validate_memo_parameters(
            token=token, request_data=request.data
        )
    except ValueError as e:
        return render_error_response(str(e))

    callback_url = request.data.get("url")
    if not callback_url:
        return render_error_response(_("callback 'url' required"))
    schemes = ["https"] if not settings.LOCAL_MODE else ["https", "http"]
    try:
        URLValidator(schemes=schemes)(request.data.get("url"))
    except ValidationError:
        return render_error_response(_("'url' must be a valid URL"))

    try:
        rci.callback(
            token=token,
            request=request,
            params={
                "id": request.data.get("id"),
                "account": token.muxed_account or token.account,
                "memo": memo,
                "memo_type": memo_type,
                "url": callback_url,
            },
        )
    except ValueError as e:
        return render_error_response(str(e), status_code=400)
    except ObjectDoesNotExist as e:
        return render_error_response(str(e), status_code=404)
    except NotImplementedError:
        return render_error_response(_("not implemented"), status_code=501)

    return Response({"success": True})


@api_view(["DELETE"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@parser_classes([MultiPartParser, FormParser, JSONParser])
@validate_sep10_token()
def delete(
    token: SEP10Token,
    request: Request,
    account: str,
) -> Response:
    if (token.muxed_account or token.account) != account:
        return render_error_response(_("account not found"), status_code=404)
    elif (
        token.memo
        and request.data.get("memo")
        and (
            str(token.memo) != str(request.data.get("memo"))
            or request.data.get("memo_type") != Transaction.MEMO_TYPES.id
        )
    ):
        return render_error_response(_("account not found"), status_code=404)
    try:
        memo, memo_type = validate_memo_parameters(
            token=token, request_data=request.data
        )
    except ValueError as e:
        return render_error_response(str(e))
    try:
        rci.delete(
            token=token,
            request=request,
            account=account,
            memo=memo,
            memo_type=memo_type,
        )
    except ObjectDoesNotExist:
        return render_error_response(_("account not found"), status_code=404)
    else:
        return Response({"status": "success"}, status=200)


@api_view(["PUT"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@parser_classes([MultiPartParser, FormParser, JSONParser])
@validate_sep10_token()
def put_verification(token: SEP10Token, request) -> Response:
    if not request.data.get("id") or not isinstance(request.data.get("id"), str):
        return render_error_response(_("bad ID value, expected str"))
    for key, value in request.data.items():
        if key == "id":
            continue
        sep9_field, verification = key.rsplit("_", 1)
        if verification != "verification" or sep9_field not in SEP_9_FIELDS:
            return render_error_response(
                _(
                    "all request attributes other than 'id' must be a SEP-9 field followed "
                    "by '_verification'"
                )
            )

    try:
        response_data = rci.put_verification(
            token=token,
            request=request,
            account=token.muxed_account or token.account,
            params=dict(request.data),
        )
    except ObjectDoesNotExist:
        return render_error_response(_("customer not found"), status_code=404)
    except ValueError as e:
        return render_error_response(str(e), status_code=400)
    except NotImplementedError:
        return render_error_response(_("not implemented"), status_code=501)

    try:
        validate_response_data(response_data)
    except ValueError:
        logger.exception(
            "An exception was raised validating PUT /customer/verification response"
        )
        return render_error_response(_("unable to process request."), status_code=500)

    return Response(response_data)


def validate_response_data(data: Dict):
    attrs = ["fields", "id", "message", "status", "provided_fields"]
    if not data:
        raise ValueError("empty response from SEP-12 get() integration")
    elif any(f not in attrs for f in data):
        raise ValueError(
            f"unexpected attribute included in GET /customer response. "
            f"Accepted attributes: {attrs}"
        )
    elif "id" in data and not isinstance(data["id"], str):
        raise ValueError("customer IDs must be strings")
    accepted_statuses = ["ACCEPTED", "PROCESSING", "NEEDS_INFO", "REJECTED"]
    if not data.get("status") or data.get("status") not in accepted_statuses:
        raise ValueError("invalid status in SEP-12 GET /customer response")
    elif (
        any(
            [
                f.get("optional") in [False, None]
                for f in data.get("fields", {}).values()
            ]
        )
        and data.get("status") == "ACCEPTED"
    ):
        raise ValueError(
            "all required 'fields' must be provided before a customer can be 'ACCEPTED'"
        )
    if data.get("fields"):
        validate_fields(data.get("fields"))
    if data.get("provided_fields"):
        validate_fields(data.get("provided_fields"), provided=True)
    if data.get("message") and not isinstance(data["message"], str):
        raise ValueError(
            "invalid message value in SEP-12 GET /customer response, should be str"
        )


def validate_id_parameters(token: SEP10Token, request_data: dict) -> Optional[Response]:
    if request_data.get("id"):
        if not isinstance(request_data.get("id"), str):
            return render_error_response(_("bad ID value, expected str"))
        elif (
            request_data.get("account")
            or request_data.get("memo")
            or request_data.get("memo_type")
        ):
            return render_error_response(
                _(
                    "requests with 'id' cannot also have 'account', 'memo', or 'memo_type'"
                )
            )
    elif request_data.get("account") and request_data.get("account") != (
        token.muxed_account or token.account
    ):
        return render_error_response(
            _("The account specified does not match authorization token"),
            status_code=403,
        )
    elif (
        token.memo
        and request_data.get("memo")
        and (
            str(token.memo) != str(request_data.get("memo"))
            or request_data.get("memo_type") != Transaction.MEMO_TYPES.id
        )
    ):
        return render_error_response(
            _("The memo specified does not match the memo ID authorized via SEP-10"),
            status_code=403,
        )


def validate_memo_parameters(
    token: SEP10Token, request_data: dict
) -> Tuple[Optional[str], Optional[str]]:
    try:
        make_memo(request_data.get("memo"), request_data.get("memo_type"))
    except (ValueError, TypeError):
        raise ValueError(_("invalid 'memo' for 'memo_type'"))
    memo = request_data.get("memo") or token.memo
    memo_type = None
    if memo:
        memo = str(memo)
        memo_type = request_data.get("memo_type") or Transaction.MEMO_TYPES.id
    return memo, memo_type


def validate_fields(fields: Dict, provided=False):
    if not isinstance(fields, Dict):
        raise ValueError(
            "invalid fields type in SEP-12 GET /customer response, should be dict"
        )
    if len(extract_sep9_fields(fields)) < len(fields):
        raise ValueError("SEP-12 GET /customer response fields must be from SEP-9")
    accepted_field_attrs = [
        "type",
        "description",
        "choices",
        "optional",
        "status",
        "error",
    ]
    accepted_types = ["string", "binary", "number", "date"]
    accepted_statuses = [
        "ACCEPTED",
        "PROCESSING",
        "REJECTED",
        "VERIFICATION_REQUIRED",
    ]
    for key, value in fields.items():
        if not set(value.keys()).issubset(set(accepted_field_attrs)):
            raise ValueError(
                f"unexpected attribute in '{key}' object in SEP-12 GET /customer response, "
                f"accepted values: {', '.join(accepted_field_attrs)}"
            )
        if not value.get("type") or value.get("type") not in accepted_types:
            raise ValueError(
                f"bad type value for '{key}' in SEP-12 GET /customer response"
            )
        elif not (
            value.get("description") and isinstance(value.get("description"), str)
        ):
            raise ValueError(
                f"bad description value for '{key}' in SEP-12 GET /customer response"
            )
        elif value.get("choices") and not isinstance(value.get("choices"), list):
            raise ValueError(
                f"bad choices value for '{key}' in SEP-12 GET /customer response"
            )
        elif value.get("optional") and not isinstance(value.get("optional"), bool):
            raise ValueError(
                f"bad optional value for '{key}' in SEP-12 GET /customer response"
            )
        elif not provided and value.get("status"):
            raise ValueError(
                f"'{key}' object in 'fields' object cannot have a 'status' property"
            )
        elif (
            provided
            and value.get("status")
            and value.get("status") not in accepted_statuses
        ):
            raise ValueError(
                f"bad field status value for '{key}' in SEP-12 GET /customer response"
            )
        elif not provided and value.get("error"):
            raise ValueError(
                f"'{key}' object in 'fields' object cannot have 'error' property"
            )
        elif (
            provided and value.get("error") and not isinstance(value.get("error"), str)
        ):
            raise ValueError(
                f"bad error value for '{key}' in SEP-12 GET /customer response"
            )



"""
Polaris-specific settings. This is not django.conf.settings.
"""
import os
import environ
from urllib.parse import urlparse
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from stellar_sdk.server import Server
from stellar_sdk.keypair import Keypair


def env_or_settings(variable, required=True, bool=False, list=False, int=False):
    try:
        if bool:
            return env.bool(variable)
        elif list:
            return env.list(variable)
        elif int:
            return env.int(variable)
        else:
            return env(variable)
    except ImproperlyConfigured as e:
        if hasattr(settings, "POLARIS_" + variable):
            return getattr(settings, "POLARIS_" + variable)
        elif required:
            raise e
        else:
            return None


env = environ.Env()
env_file = os.path.join(getattr(settings, "BASE_DIR", ""), ".env")
if os.path.exists(env_file):
    env.read_env(env_file)
elif hasattr(settings, "POLARIS_ENV_PATH"):
    if os.path.exists(settings.POLARIS_ENV_PATH):
        env.read_env(settings.POLARIS_ENV_PATH)
    else:
        raise ImproperlyConfigured(
            f"Could not find env file at {settings.POLARIS_ENV_PATH}"
        )

accepted_seps = [
    "sep-1",
    "sep-6",
    "sep-6",
    "sep-10",
    "sep-12",
    "sep-24",
    "sep-31",
    "sep-38",
]
ACTIVE_SEPS = env_or_settings("ACTIVE_SEPS", list=True)
for i, sep in enumerate(ACTIVE_SEPS):
    if sep.lower() not in accepted_seps:
        raise ImproperlyConfigured(
            f"Unrecognized value in ACTIVE_SEPS list: {sep}; Accepted values: {accepted_seps}"
        )
    ACTIVE_SEPS[i] = sep.lower()

SIGNING_SEED, SIGNING_KEY = None, None
if "sep-10" in ACTIVE_SEPS:
    SIGNING_SEED = env_or_settings("SIGNING_SEED")
    try:
        SIGNING_KEY = Keypair.from_secret(SIGNING_SEED).public_key
    except ValueError:
        raise ImproperlyConfigured("Invalid SIGNING_SEED")

SERVER_JWT_KEY = None
if any(sep in ACTIVE_SEPS for sep in ["sep-10", "sep-24"]):
    SERVER_JWT_KEY = env_or_settings("SERVER_JWT_KEY")

STELLAR_NETWORK_PASSPHRASE = (
    env_or_settings("STELLAR_NETWORK_PASSPHRASE", required=False)
    or "Test SDF Network ; September 2015"
)

HORIZON_URI = (
    env_or_settings("HORIZON_URI", required=False)
    or "https://horizon-testnet.stellar.org"
)
if not HORIZON_URI.startswith("http"):
    raise ImproperlyConfigured("HORIZON_URI must include a protocol (http or https)")
HORIZON_SERVER = Server(horizon_url=HORIZON_URI)

LOCAL_MODE = env_or_settings("LOCAL_MODE", bool=True, required=False) or False

HOST_URL = env_or_settings("HOST_URL")
if not HOST_URL.startswith("http"):
    raise ImproperlyConfigured("HOST_URL must include a protocol (http or https)")
elif LOCAL_MODE and HOST_URL.startswith("https"):
    raise ImproperlyConfigured("HOST_URL uses HTTPS but LOCAL_MODE only supports HTTP")
elif not LOCAL_MODE and not HOST_URL.startswith("https"):
    raise ImproperlyConfigured("HOST_URL uses HTTP but LOCAL_MODE is off")

SEP10_HOME_DOMAINS = env_or_settings(
    "SEP10_HOME_DOMAINS", list=True, required=False
) or [urlparse(HOST_URL).netloc]
if any(d.startswith("http") for d in SEP10_HOME_DOMAINS):
    raise ImproperlyConfigured("SEP10_HOME_DOMAINS must only be hostnames")

MAX_TRANSACTION_FEE_STROOPS = env_or_settings(
    "MAX_TRANSACTION_FEE_STROOPS", int=True, required=False
)

CALLBACK_REQUEST_TIMEOUT = (
    env_or_settings("CALLBACK_REQUEST_TIMEOUT", int=True, required=False) or 3
)
CALLBACK_REQUEST_DOMAIN_DENYLIST = (
    env_or_settings("CALLBACK_REQUEST_DOMAIN_DENYLIST", list=True, required=False) or []
)

SEP6_USE_MORE_INFO_URL = (
    env_or_settings("SEP6_USE_MORE_INFO_URL", bool=True, required=False) or False
)

SEP10_CLIENT_ATTRIBUTION_REQUIRED = env_or_settings(
    "SEP10_CLIENT_ATTRIBUTION_REQUIRED", bool=True, required=False
)
SEP10_CLIENT_ATTRIBUTION_REQUEST_TIMEOUT = (
    env_or_settings(
        "SEP10_CLIENT_ATTRIBUTION_REQUEST_TIMEOUT", int=True, required=False
    )
    or 3
)
SEP10_CLIENT_ATTRIBUTION_ALLOWLIST = env_or_settings(
    "SEP10_CLIENT_ATTRIBUTION_ALLOWLIST", list=True, required=False
)
SEP10_CLIENT_ATTRIBUTION_DENYLIST = env_or_settings(
    "SEP10_CLIENT_ATTRIBUTION_DENYLIST", list=True, required=False
)

ADDITIVE_FEES_ENABLED = (
    env_or_settings("ADDITIVE_FEES_ENABLED", bool=True, required=False) or False
)

# Constants
OPERATION_DEPOSIT = "deposit"
OPERATION_WITHDRAWAL = "withdraw"
OPERATION_SEND = "send"
ACCOUNT_STARTING_BALANCE = str(2.01)
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

INTERACTIVE_JWT_EXPIRATION = (
    env_or_settings("INTERACTIVE_JWT_EXPIRATION", int=True, required=False) or 30
)
if INTERACTIVE_JWT_EXPIRATION <= 0:
    raise ImproperlyConfigured("INTERACTIVE_JWT_EXPIRATION must be positive")



from decimal import Decimal, DecimalException
from typing import Dict, Tuple

from django.utils.translation import gettext as _
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.decorators import api_view, renderer_classes, parser_classes
from rest_framework.parsers import JSONParser, MultiPartParser, FormParser
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer
from stellar_sdk import Keypair
from stellar_sdk.strkey import StrKey
from stellar_sdk.exceptions import (
    MemoInvalidException,
    Ed25519PublicKeyInvalidError,
    MuxedEd25519AccountInvalidError,
    ValueError as StellarSdkValueError,
)

from polaris import settings
from polaris.utils import (
    getLogger,
    render_error_response,
    create_transaction_id,
    extract_sep9_fields,
    make_memo,
    get_quote_and_offchain_destination_asset,
    validate_account_and_memo,
)
from polaris.sep6.utils import validate_403_response
from polaris.sep10.token import SEP10Token
from polaris.sep10.utils import validate_sep10_token
from polaris.shared.endpoints import SEP6_MORE_INFO_PATH
from polaris.locale.utils import (
    activate_lang_for_request,
    validate_or_use_default_language,
)
from polaris.models import Asset, Transaction, Quote
from polaris.integrations import (
    registered_withdrawal_integration as rwi,
    registered_fee_func,
    calculate_fee,
    registered_custody_integration as rci,
)

logger = getLogger(__name__)


@api_view(["GET"])
@parser_classes([MultiPartParser, FormParser, JSONParser])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@validate_sep10_token()
def withdraw(token: SEP10Token, request: Request) -> Response:
    return withdraw_logic(token=token, request=request, exchange=False)


@api_view(["GET"])
@parser_classes([MultiPartParser, FormParser, JSONParser])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@validate_sep10_token()
def withdraw_exchange(token: SEP10Token, request: Request) -> Response:
    return withdraw_logic(token=token, request=request, exchange=True)


def withdraw_logic(token: SEP10Token, request: Request, exchange: bool):
    args = parse_request_args(token, request, exchange)
    if "error" in args:
        return args["error"]

    transaction_id = create_transaction_id()
    transaction = Transaction(
        id=transaction_id,
        stellar_account=token.account,
        muxed_account=token.muxed_account,
        account_memo=token.memo,
        asset=args["source_asset" if exchange else "asset"],
        amount_in=args["amount"],
        amount_expected=args["amount"],
        quote=args["quote"],
        kind=getattr(
            Transaction.KIND, "withdrawal-exchange" if exchange else "withdrawal"
        ),
        status=Transaction.STATUS.pending_user_transfer_start,
        protocol=Transaction.PROTOCOL.sep6,
        more_info_url=request.build_absolute_uri(
            f"{SEP6_MORE_INFO_PATH}?id={transaction_id}"
        ),
        on_change_callback=args["on_change_callback"],
        client_domain=token.client_domain,
        from_address=args.get("account"),
    )

    try:
        integration_response = rwi.process_sep6_request(
            token=token, request=request, params=args, transaction=transaction
        )
    except ValueError as e:
        return render_error_response(str(e))
    try:
        response_attrs, status_code = validate_response(
            token=token,
            request=request,
            args=args,
            integration_response=integration_response,
            transaction=transaction,
            exchange=exchange,
        )
    except ValueError as e:
        logger.error(str(e))
        return render_error_response(
            _("unable to process the request"), status_code=500
        )

    if status_code != 200:
        if Transaction.objects.filter(id=transaction.id).exists():
            logger.error("Do not save transaction objects for invalid SEP-6 requests")
            return render_error_response(
                _("unable to process the request"), status_code=500
            )
        return Response(response_attrs, status=status_code)

    try:
        receiving_account, memo_str, memo_type = validate_account_and_memo(
            *rci.get_receiving_account_and_memo(
                request=request, transaction=transaction
            )
        )
    except ValueError:
        logger.exception(
            "CustodyIntegration.get_receiving_account_and_memo() returned invalid values"
        )
        return render_error_response(
            _("unable to process the request"), status_code=500
        )
    transaction.receiving_anchor_account = receiving_account
    transaction.memo = memo_str
    transaction.memo_type = memo_type
    response_attrs["memo"] = transaction.memo
    response_attrs["memo_type"] = transaction.memo_type
    logger.info(f"Created withdraw transaction {transaction.id}")
    if exchange:
        transaction.quote.save()
    transaction.save()

    return Response(
        {
            "id": transaction.id,
            "account_id": transaction.receiving_anchor_account,
            **response_attrs,
        },
        status=status_code,
    )


def parse_request_args(
    token: SEP10Token, request: Request, exchange: bool = False
) -> Dict:
    lang = validate_or_use_default_language(request.GET.get("lang"))
    activate_lang_for_request(lang)

    account = request.GET.get("account")
    if account and account.startswith("M"):
        try:
            StrKey.decode_muxed_account(account)
        except (MuxedEd25519AccountInvalidError, StellarSdkValueError):
            return {"error": render_error_response(_("invalid 'account'"))}
    elif account:
        try:
            Keypair.from_public_key(account)
        except Ed25519PublicKeyInvalidError:
            return {"error": render_error_response(_("invalid 'account'"))}

    if exchange:
        if not request.GET.get("source_asset"):
            return {
                "error": render_error_response(_("'destination_asset' is required"))
            }
        parts = request.GET.get("source_asset").split(":")
        if len(parts) != 3 or parts[0] != "stellar":
            return {"error": render_error_response(_("invalid 'source_asset'"))}
        asset_query_args = {"code": parts[1], "issuer": parts[2]}
    else:
        asset_query_args = {"code": request.GET.get("asset_code")}

    asset = Asset.objects.filter(
        sep6_enabled=True, withdrawal_enabled=True, **asset_query_args
    ).first()
    if not asset:
        return {
            "error": render_error_response(_("invalid 'asset_code' or 'source_asset'"))
        }

    memo_type = request.GET.get("memo_type")
    if memo_type and memo_type not in Transaction.MEMO_TYPES:
        return {"error": render_error_response(_("invalid 'memo_type'"))}

    try:
        make_memo(request.GET.get("memo"), memo_type)
    except (ValueError, TypeError, MemoInvalidException):
        return {"error": render_error_response(_("invalid 'memo' for 'memo_type'"))}

    if not request.GET.get("type"):
        return {"error": render_error_response(_("'type' is required"))}
    if not request.GET.get("dest"):
        return {"error": render_error_response(_("'dest' is required"))}

    on_change_callback = request.GET.get("on_change_callback")
    if on_change_callback and on_change_callback.lower() != "postmessage":
        schemes = ["https"] if not settings.LOCAL_MODE else ["https", "http"]
        try:
            URLValidator(schemes=schemes)(on_change_callback)
        except ValidationError:
            return {"error": render_error_response(_("invalid callback URL provided"))}
        if any(
            domain in on_change_callback
            for domain in settings.CALLBACK_REQUEST_DOMAIN_DENYLIST
        ):
            on_change_callback = None

    amount = request.GET.get("amount")
    if exchange and not amount:
        return {"error": render_error_response(_("'amount' is required"))}

    if amount:
        try:
            amount = round(Decimal(amount), asset.significant_decimals)
        except DecimalException:
            return {"error": render_error_response(_("invalid 'amount'"))}
        min_amount = round(asset.withdrawal_min_amount, asset.significant_decimals)
        max_amount = round(asset.withdrawal_max_amount, asset.significant_decimals)
        if not (min_amount <= amount <= max_amount):
            return {
                "error": render_error_response(
                    _("'amount' must be within [%s, %s]") % (min_amount, min_amount)
                )
            }

    try:
        quote, destination_asset = get_quote_and_offchain_destination_asset(
            token=token,
            quote_id=request.GET.get("quote_id"),
            destination_asset_str=request.GET.get("destination_asset"),
            asset=asset,
            amount=amount,
        )
    except ValueError as e:
        return {"error": render_error_response(str(e))}

    if (
        quote
        and quote.type == Quote.TYPE.firm
        and Transaction.objects.filter(quote=quote).exists()
    ):
        return {
            "error": render_error_response(
                _("quote has already been used in a transaction")
            )
        }

    args = {
        "account": request.GET.get("account"),
        "source_asset" if exchange else "asset": asset,
        "memo_type": memo_type,
        "memo": request.GET.get("memo"),
        "lang": lang,
        "type": request.GET.get("type"),
        "dest": request.GET.get("dest"),
        "dest_extra": request.GET.get("dest_extra"),
        "on_change_callback": on_change_callback,
        "amount": amount,
        "country_code": request.GET.get("country_code"),
        "quote": quote,
        "destination_asset": destination_asset,
        **extract_sep9_fields(request.GET),
    }

    # add remaining extra params, it's on the anchor to validate them
    for param, value in request.GET.items():
        if param not in args:
            args[param] = value

    return args


def validate_response(
    token: SEP10Token,
    request: Request,
    args: Dict,
    integration_response: Dict,
    transaction: Transaction,
    exchange: bool = False,
) -> Tuple[Dict, int]:
    if "type" in integration_response:
        return (
            validate_403_response(token, request, integration_response, transaction),
            403,
        )

    asset = args["source_asset" if exchange else "asset"]
    response_attrs = {}
    if "min_amount" in integration_response:
        response_attrs["min_amount"] = integration_response["min_amount"]
        if type(integration_response["min_amount"]) not in [Decimal, int, float]:
            raise ValueError(
                "Invalid 'min_amount' type returned from process_sep6_request()"
            )
        elif integration_response["min_amount"] < 0:
            raise ValueError(
                "Invalid 'min_amount' returned from process_sep6_request()"
            )
    elif (
        transaction.asset.withdrawal_min_amount
        > getattr(Asset, "_meta").get_field("withdrawal_min_amount").default
    ):
        response_attrs["min_amount"] = round(
            transaction.asset.withdrawal_min_amount,
            transaction.asset.significant_decimals,
        )
    if "max_amount" in integration_response:
        if type(integration_response["max_amount"]) not in [Decimal, int, float]:
            raise ValueError(
                "Invalid 'max_amount' type returned from process_sep6_request()"
            )
        elif integration_response["max_amount"] < 0:
            raise ValueError(
                "Invalid 'max_amount' returned from process_sep6_request()"
            )
        response_attrs["max_amount"] = integration_response["max_amount"]
    elif (
        transaction.asset.withdrawal_max_amount
        < getattr(Asset, "_meta").get_field("withdrawal_max_amount").default
    ):
        response_attrs["max_amount"] = round(
            transaction.asset.withdrawal_max_amount,
            transaction.asset.significant_decimals,
        )

    if "fee_fixed" in integration_response or "fee_percent" in integration_response:
        if "fee_fixed" in integration_response:
            response_attrs["fee_fixed"] = integration_response["fee_fixed"]
        if "fee_percent" in integration_response:
            response_attrs["fee_percent"] = integration_response["fee_percent"]
    elif calculate_fee == registered_fee_func and not exchange:
        # return the fixed and percentage fee rates if the registered fee function
        # has not been implemented AND the request was not for the `/withdraw-exchange`
        # endpoint.
        if asset.withdrawal_fee_fixed is not None:
            response_attrs["fee_fixed"] = round(
                asset.withdrawal_fee_fixed, asset.significant_decimals
            )
        if asset.withdrawal_fee_percent is not None:
            response_attrs["fee_percent"] = round(
                asset.withdrawal_fee_percent, asset.significant_decimals
            )

    if "extra_info" in integration_response:
        if not isinstance(integration_response["extra_info"], dict):
            raise ValueError("invalid 'extra_info' returned from integration")
        response_attrs["extra_info"] = integration_response["extra_info"]

    return response_attrs, 200



# Generated by Django 2.2.18 on 2021-02-26 00:53

from django.db import migrations
import polaris.models


def encrypt_channel_seeds(apps, _):
    Transaction = apps.get_model("polaris", "Transaction")
    channel_seed_txs = Transaction.objects.filter(channel_seed__isnull=False).all()
    for tx in channel_seed_txs:
        tx.channel_seed = polaris.models.EncryptedTextField.encrypt(tx.channel_seed)
        tx.save()


def decrypt_channel_seeds(apps, _):
    Transaction = apps.get_model("polaris", "Transaction")
    channel_seed_txs = Transaction.objects.filter(channel_seed__isnull=False).all()
    for tx in channel_seed_txs:
        tx.channel_seed = polaris.models.EncryptedTextField.decrypt(tx.channel_seed)
        tx.save()


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0004_auto_20201120_1541"),
    ]

    operations = [
        migrations.RunPython(encrypt_channel_seeds, decrypt_channel_seeds),
        migrations.AlterField(
            model_name="transaction",
            name="channel_seed",
            field=polaris.models.EncryptedTextField(blank=True, null=True),
        ),
    ]



import enum
from typing import List, Union, Optional, Tuple

from stellar_sdk import Server, Keypair, TransactionBuilder, TransactionEnvelope
from stellar_sdk.exceptions import (
    ConnectionError,
    BadRequestError,
    BadResponseError,
)
from rest_framework.request import Request
from stellar_sdk.sep.ed25519_public_key_signer import Ed25519PublicKeySigner
from stellar_sdk.sep.stellar_web_authentication import _verify_transaction_signatures

from polaris.exceptions import (
    TransactionSubmissionPending,
    TransactionSubmissionBlocked,
    TransactionSubmissionFailed,
)
from polaris.models import Transaction, Asset
from polaris.utils import (
    getLogger,
    memo_hex_to_base64,
    create_deposit_envelope,
    get_account_obj,
)
from polaris import settings


logger = getLogger(__name__)


class CustodyIntegration:
    """
    The base class for supporting third party custody service providers.
    """

    def get_distribution_account(self, asset: Asset) -> str:
        """
        Return the Stellar account used to receive payments of `asset`. This
        method is a replacement for the ``Asset.distribution_account`` property
        which is derived from the ``Asset.distribution_seed`` database column.

        This means that the same distribution account should always be returned
        for the same asset. **Do not implement this method if your custody service
        provider does not support using the same Stellar account for all incoming
        payments of an asset.** Some custody service providers provide a Stellar
        account and memo to use as the destination of an incoming payment on a
        per-transaction basis, with no guaranteed consistency for the Stellar
        account provided.

        The ``watch_transactions`` command assumes this method is implemented.
        If this method is not implemented, another appraoch to detecting and
        matching incoming payments must be used.

        :param asset: the asset sent in payments to the returned Stellar account
        """
        raise NotImplementedError()

    def get_receiving_account_and_memo(
        self, request: Request, transaction: Transaction
    ) -> Tuple[str, str]:
        """
        Return the Stellar account that the client should use as the destination
        of the payment transaction and the string representation of the memo the
        client should attach to the transaction. Polaris will save these to
        `transaction` and return them in API responses when necessary.

        This method is only called once for a given transaction.

        The memo returned _must_ be unique to the transaction, since the anchor
        is expected to match the database record represented by `transaction` with
        the on-chain transaction submitted.

        This function differs from ``get_distribution_account()`` by allowing the
        anchor to return any Stellar account that can be used to receive a payment.
        This is ideal when the account used is determined by a custody service
        provider that does not guarantee that the account provided will be the
        account provided for future transactions.

        :param request: the request that initiated the call to this function
        :param transaction: the transaction that will be processed when a payment
            using the account and memo returned is received
        """
        raise NotImplementedError()

    def submit_deposit_transaction(
        self, transaction: Transaction, has_trustline: bool = True
    ) -> str:
        """
        Submit the transaction to the Stellar network using the anchor's custody
        service provider and return the hash of the transaction once it is included
        in a ledger

        **Claimable Balances**

        If ``self.claimable_balances_supported`` is ``True``, Polaris may call
        this method when the destination account does not yet have a trustline
        to ``Transaction.asset``. In this case, the anchor should send the
        deposit as a claimable balance instead of a payment or path payment. Use
        the `has_trustline` parameter to determine which operations to use.

        If ``self.claimable_balances_supported`` is ``False``, this method will only
        be called when the destination account exists and has a trustline to
        ``Transaction.asset``.

        **Handling Non-Success Cases**

        Raise a ``TransactionSubmissionPending`` exception if the call(s) made to
        initiate the transaction submission process did not result in the immediate
        inclusion of the transaction in a ledger. Polaris will simply call this
        function again with the same transaction parameters unless a SIGINT or
        SIGTERM signal has been sent to the process, in which case Polaris will
        save ``Transaction.submission_status`` as ``SubmissionStatus.PENDING`` and
        exit. When the process starts up, Polaris will retrieve the currently
        pending transaction from the database and pass it to this function again.

        Raise a ``TransactionSubmissionBlocked`` exception if the transaction is not
        yet ready to be submitted. A transaction may be blocked for a number of
        reasons excluding Polaris' default checks for account and trustline
        existence. Polaris will update ``Transaction.submission_status`` to
        ``SubmissionStatus.BLOCKED`` and expected the anchor to update this field
        to ``SubmissionStatus.UNBLOCKED`` once Polaris should attempt submission again.

        :param transaction: the ``Transaction`` object representing the Stellar
            transaction that should be submitted to the network
        :param has_trustline: whether or not the destination account has a trustline
            for the requested asset
        """
        raise NotImplementedError()

    def create_destination_account(self, transaction: Transaction) -> str:
        """
        Submit a transaction using the anchor's custody service provider to fund
        the Stellar account address saved to ``Transaction.to_address`` and return
        the hash of the transaction once it is included in a ledger

        If ``self.account_creation_supported`` is ``False`` Polaris will never call
        this method. However, Polaris will instead check if destination accounts
        exist when a request for deposit is made and will return a
        422 Unprocessable Entity response if it does not.

        It is highly recommended to support creating destination accounts.

        **Handling Non-Success Cases**

        Raise a ``TransactionSubmissionPending`` exception if the call(s) made to
        initiate the transaction submission process did not result in the immediate
        inclusion of the transaction in a ledger. Polaris will simply call this
        function again with the same transaction parameters unless a SIGINT or
        SIGTERM signal has been sent to the process, in which case Polaris will
        save ``Transaction.submission_status`` as ``SubmissionStatus.PENDING`` and
        exit. When the process starts up, Polaris will retrieve the currently
        pending transaction from the database and pass it to this function again.

        Raise a ``TransactionSubmissionBlocked`` exception if the transaction is not
        yet ready to be submitted. A transaction may be blocked for a number of
        reasons excluding Polaris' default checks for account and trustline
        existence. Polaris will update ``Transaction.submission_status`` to
        ``SubmissionStatus.BLOCKED`` and expected the anchor to update this field
        to ``SubmissionStatus.UNBLOCKED`` once Polaris should attempt submission again.

        :param transaction: the transaction for which the destination account must
            be funded
        """
        raise NotImplementedError()

    @property
    def claimable_balances_supported(self) -> bool:
        """
        Return ``True`` if the custody service provider supports sending deposit
        payments in the form of claimable balances, ``False`` otherwise.
        """
        raise NotImplementedError()

    @property
    def account_creation_supported(self) -> bool:
        """
        Return ``True`` if the custody service provider supports funding Stellar
        accounts not custodied by the provider, ``False`` otherwise.
        """
        raise NotImplementedError()


class SelfCustodyIntegration(CustodyIntegration):
    """
    The default custody class used if no other custody class is registered.
    Assumes a Stellar account secret key is saved to ``Asset.distribution_seed``
    for every anchored asset.
    """

    class TransactionType(enum.Enum):
        CREATE_DESTINATION_ACCOUNT = 1
        SEND_DEPOSIT_AMOUNT = 2

    @property
    def claimable_balances_supported(self):
        """
        Polaris supports sending funds for deposit transactions as claimable balances.
        """
        return True

    @property
    def account_creation_supported(self):
        """
        Polaris supports funding destination accounts for deposit transactions.
        """
        return True

    def get_distribution_account(self, asset: Asset) -> str:
        """
        Returns ``Asset.distribution_account``
        """
        return asset.distribution_account

    def get_receiving_account_and_memo(
        self, request: Request, transaction: Transaction
    ):
        """
        Returns the distribution account used for the transaction's asset and a
        hash memo derived from the transaction's ID.
        """
        padded_hex_memo = "0" * (64 - len(transaction.id.hex)) + transaction.id.hex
        return (
            transaction.asset.distribution_account,
            memo_hex_to_base64(padded_hex_memo),
        )

    def submit_deposit_transaction(
        self, transaction: Transaction, has_trustline: bool = True
    ) -> str:
        return self._submit_transaction(
            transaction, self.TransactionType.SEND_DEPOSIT_AMOUNT, has_trustline
        )

    def create_destination_account(self, transaction: Transaction) -> dict:
        return self._submit_transaction(
            transaction, self.TransactionType.CREATE_DESTINATION_ACCOUNT
        )

    def _submit_transaction(
        self, transaction, type: TransactionType, has_trustline: Optional[bool] = None
    ):
        try:
            requires_add_sigs = self._requires_additional_signatures(transaction)
        except ConnectionError as e:
            # Unable to fetch the information necessary to determine whether
            # the transaction requires additional signatures. We're going to
            # assume this issue is temporary and instruct Polaris retry
            # submitting this transaction.
            raise TransactionSubmissionPending(f"ConnectionError: {str(e)}")
        with Server(horizon_url=settings.HORIZON_URI) as server:
            if requires_add_sigs:
                # The transaction requires signatures in addition to the
                # signatures Polaris can add directly, namely the signatures
                # from the asset's distribution account and the channel account
                # provided by the anchor. So, we indicate to Polaris that it
                # should wait until the anchor signals that this transaction is
                # ready for submission.
                self._save_as_pending_signatures(
                    transaction, type, server, has_trustline
                )
                raise TransactionSubmissionBlocked(
                    "non-master distribution account signatures need to be " "collected"
                )
            transaction_hash = None
            while not transaction_hash:
                if transaction.envelope_xdr:
                    envelope = transaction.envelope_xdr
                else:
                    if type == self.TransactionType.SEND_DEPOSIT_AMOUNT:
                        envelope_obj = self._generate_deposit_transaction_envelope(
                            transaction, has_trustline, server
                        )
                    else:
                        envelope_obj = (
                            self._generate_create_account_transaction_envelope(
                                transaction,
                                server,
                            )
                        )
                    envelope = self._sign_and_save_transaction_envelope(
                        transaction, envelope_obj, [transaction.asset.distribution_seed]
                    )
                try:
                    response = server.submit_transaction(envelope)
                except (BadResponseError, ConnectionError) as e:
                    if isinstance(e, BadResponseError):
                        exception_msg = f"BadResponseError ({e.status}): {e.message}"
                    else:
                        exception_msg = str(e)
                    raise TransactionSubmissionPending(exception_msg)
                except BadRequestError as e:
                    if e.status != 400:
                        raise TransactionSubmissionFailed(
                            f"unexpected status code: {e.status}"
                        )
                    # We assume we constructed the transaction properly. This rules
                    # out several possible transaction error codes.
                    #
                    # The only transaction error code we will attempt to recover from
                    # is tx_insufficient_fee, which indicates the network is in surge
                    # pricing mode, therefore we need to retry submitting.
                    #
                    # All other transaction error codes are considered the fault of
                    # anchor (ex. insufficient balances, not adding required signers)
                    tx_error_code = e.extras.get("result_codes", {}).get("transaction")
                    exception_msg = f"BadRequestError ({e.status}): {e.message}"
                    if tx_error_code == "tx_insufficient_fee":
                        raise TransactionSubmissionPending(exception_msg)
                    else:
                        # unexpected transaction error code.
                        raise TransactionSubmissionFailed(exception_msg)
                return response["hash"]

    @staticmethod
    def _generate_deposit_transaction_envelope(
        transaction: Transaction, has_trustline: bool, server: Server
    ) -> TransactionEnvelope:
        if transaction.channel_account:
            source_account, _ = get_account_obj(
                Keypair.from_public_key(transaction.channel_account)
            )
        else:
            source_account, _ = get_account_obj(
                Keypair.from_public_key(transaction.asset.distribution_account)
            )
        return create_deposit_envelope(
            transaction=transaction,
            source_account=source_account,
            use_claimable_balance=not has_trustline,
            base_fee=(settings.MAX_TRANSACTION_FEE_STROOPS or server.fetch_base_fee()),
        )

    @staticmethod
    def _generate_create_account_transaction_envelope(
        transaction: Transaction, server: Server
    ) -> TransactionEnvelope:
        if transaction.channel_account:
            source_account, _ = get_account_obj(
                Keypair.from_public_key(transaction.channel_account)
            )
        else:
            source_account, _ = get_account_obj(
                Keypair.from_public_key(transaction.asset.distribution_account)
            )
        builder = TransactionBuilder(
            source_account=source_account,
            network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
            # this transaction contains one operation so base_fee will be multiplied by 1
            base_fee=(settings.MAX_TRANSACTION_FEE_STROOPS or server.fetch_base_fee()),
        )
        return builder.append_create_account_op(
            source=transaction.asset.distribution_account,
            destination=transaction.to_address,
            starting_balance=settings.ACCOUNT_STARTING_BALANCE,
        ).build()

    @staticmethod
    def _requires_additional_signatures(transaction: Transaction) -> bool:
        """
        Checks if the transaction is using a distribution account that requires
        non-master signers and if so, checks if the transaction envelope has the
        signatures from enough non-master signers.

        Note that this function does not check for the presence of a signature
        from the transaction's channel account. It is assumed Polaris adds the
        signature from a channel account when the transaction envelope is crafted.
        """
        master_signer = transaction.asset.get_distribution_account_master_signer()
        thresholds = transaction.asset.get_distribution_account_thresholds()
        is_multisig_account = (
            not master_signer
            or master_signer["weight"] == 0
            or master_signer["weight"] < thresholds["med_threshold"]
        )
        if not is_multisig_account:
            return False
        if not transaction.envelope_xdr:
            return True
        possible_signers = []
        for signer_json in transaction.asset.get_distribution_account_signers():
            possible_signers.append(
                Ed25519PublicKeySigner(
                    account_id=signer_json["key"], weight=signer_json["weight"]
                )
            )
        envelope = TransactionEnvelope.from_xdr(
            transaction.envelope_xdr, settings.STELLAR_NETWORK_PASSPHRASE
        )
        signers = _verify_transaction_signatures(envelope, possible_signers)
        return sum(s.weight for s in signers) < thresholds["med_threshold"]

    def _save_as_pending_signatures(
        self,
        transaction: Transaction,
        type: TransactionType,
        server: Server,
        has_trustline: Optional[bool] = None,
    ):
        """
        Saves the transaction in a state such that the anchor can query the transaction
        from the database and sign the transaction envelope with an abitrary number of
        additional signers required to submit the transaction.

        Currently, Polaris assumes the anchor has a pool of channel accounts that can be
        used as the source account on this transaction so that it can be submitted at a
        later point in time without failing with a bad sequence number error.

        NOTE: Polaris will drop support for multisig transactions in v3.0.
        """
        from polaris.integrations import registered_deposit_integration as rdi

        if not transaction.channel_account:
            # we haven't called this yet
            rdi.create_channel_account(transaction)
        if not transaction.channel_account:
            # maybe the anchor used a Transaction.objects.update() query and
            # channel_account isn't saved to the object.
            # refresh but don't remove foriegn key objects previously queried
            asset = transaction.asset
            quote = transaction.quote
            transaction.refresh_from_db()
            transaction.asset = asset
            transaction.quote = quote
        if not transaction.channel_account:
            # the anchor didn't properly implement create_channel_account
            raise TransactionSubmissionFailed(
                "DepositIntegration.create_channel_account() did not save a "
                "secret key to Transaction.channel_seed"
            )
        if type == self.TransactionType.SEND_DEPOSIT_AMOUNT:
            envelope = self._generate_deposit_transaction_envelope(
                transaction=transaction, has_trustline=has_trustline, server=server
            )
        else:
            envelope = self._generate_create_account_transaction_envelope(
                transaction=transaction, server=server
            )
        transaction.pending_signatures = True
        self._sign_and_save_transaction_envelope(
            transaction=transaction,
            envelope=envelope,
            signers=[transaction.asset.distribution_seed, transaction.channel_seed],
        )

    @staticmethod
    def _sign_and_save_transaction_envelope(
        transaction: Transaction,
        envelope: TransactionEnvelope,
        signers: List[Union[Keypair, str]],
    ) -> str:
        for signer in signers:
            envelope.sign(signer)
        transaction.envelope_xdr = envelope.to_xdr()
        transaction.save()
        return transaction.envelope_xdr


registered_custody_integration = SelfCustodyIntegration()



# Generated by Django 2.2.16 on 2020-09-21 21:35

from django.db import migrations, models
import polaris.models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="envelope_xdr",
            field=models.TextField(
                null=True, validators=[polaris.models.deserialize], blank=True
            ),
        ),
        migrations.AddField(
            model_name="transaction",
            name="pending_signatures",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="transaction",
            name="channel_seed",
            field=models.TextField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name="asset",
            name="distribution_account_signers",
            field=models.TextField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name="asset",
            name="distribution_account_thresholds",
            field=models.TextField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name="asset",
            name="distribution_account_master_signer",
            field=models.TextField(null=True, blank=True),
        ),
    ]



"""This module defines helpers for various endpoints."""
import base64
import json
import codecs
import time
from urllib.parse import urlparse
import uuid
from datetime import datetime, timezone
from logging import getLogger
from typing import Optional, Union, Tuple, Dict
from decimal import Decimal

import aiohttp
from django.utils.translation import gettext as _
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext
from rest_framework import status
from rest_framework.response import Response
from stellar_sdk import (
    TransactionBuilder,
    TransactionEnvelope,
    Asset as StellarAsset,
    Claimant,
    TextMemo,
    IdMemo,
    HashMemo,
    Keypair,
)
from stellar_sdk.exceptions import (
    NotFoundError,
    Ed25519PublicKeyInvalidError,
    MemoInvalidException,
)
from stellar_sdk.account import Account
from stellar_sdk import Memo
from requests import Response as RequestsResponse, RequestException, post
from aiohttp import ClientResponse

from polaris import settings
from polaris.models import Transaction, Asset, Quote, OffChainAsset, ExchangePair
from polaris.sep10.token import SEP10Token
from polaris.sep38.utils import asset_id_to_kwargs
from polaris.shared.serializers import TransactionSerializer


logger = getLogger(__name__)


def render_error_response(
    description: str,
    status_code: int = status.HTTP_400_BAD_REQUEST,
    as_html: bool = False,
) -> Response:
    """
    Renders an error response in Django.

    Currently supports HTML or JSON responses.
    """
    resp_data = {"data": {"error": description}, "status": status_code}
    if as_html:
        resp_data["data"]["status_code"] = str(status_code)
        resp_data["template_name"] = "polaris/error.html"
    return Response(**resp_data)


def memo_hex_to_base64(memo):
    """
    Formats a hex memo, as in the Transaction model, to match
    the base64 Horizon response.
    """
    return (codecs.encode(codecs.decode(memo, "hex"), "base64").decode("utf-8")).strip()


def memo_base64_to_hex(memo):
    return (
        codecs.encode(codecs.decode(memo.encode(), "base64"), "hex").decode("utf-8")
    ).strip()


def create_transaction_id():
    """Creates a unique UUID for a Transaction, via checking existing entries."""
    while True:
        transaction_id = uuid.uuid4()
        if not Transaction.objects.filter(id=transaction_id).exists():
            break
    return transaction_id


def verify_valid_asset_operation(
    asset, amount, op_type, as_html=False
) -> Optional[Response]:
    enabled = getattr(asset, f"{op_type}_enabled")
    min_amount = getattr(asset, f"{op_type}_min_amount")
    max_amount = getattr(asset, f"{op_type}_max_amount")
    if not enabled:
        return render_error_response(
            gettext("the specified operation is not available for '%s'") % asset.code,
            as_html=as_html,
        )
    elif not (min_amount <= amount <= max_amount):
        return render_error_response(
            gettext("Asset amount must be within bounds [%(min)s, %(max)s]")
            % {
                "min": round(min_amount, asset.significant_decimals),
                "max": round(max_amount, asset.significant_decimals),
            },
            as_html=as_html,
        )


def load_account(resp):
    sequence = int(resp["sequence"])
    account = Account(account=resp["account_id"], sequence=sequence, raw_data=resp)
    return account


def get_account_obj(kp):
    try:
        json_resp = (
            settings.HORIZON_SERVER.accounts()
            .account_id(account_id=kp.public_key)
            .call()
        )
    except NotFoundError:
        raise RuntimeError(f"account {kp.public_key} does not exist")
    else:
        return load_account(json_resp), json_resp


async def get_account_obj_async(kp, server):
    try:
        json_resp = await server.accounts().account_id(account_id=kp.public_key).call()
    except NotFoundError:
        raise RuntimeError(f"account {kp.public_key} does not exist")
    else:
        return load_account(json_resp), json_resp


def is_pending_trust(transaction, json_resp):
    pending_trust = True
    for balance in json_resp["balances"]:
        if balance.get("asset_type") in ["native", "liquidity_pool_shares"]:
            continue
        asset_code = balance["asset_code"]
        asset_issuer = balance["asset_issuer"]
        if (
            transaction.asset.code == asset_code
            and transaction.asset.issuer == asset_issuer
        ):
            pending_trust = False
            break
    return pending_trust


def memo_str(memo: Optional[Memo]) -> Tuple[Optional[str], Optional[str]]:
    if not memo:
        return memo, None
    if isinstance(memo, IdMemo):
        return str(memo.memo_id), Transaction.MEMO_TYPES.id
    elif isinstance(memo, HashMemo):
        return memo_hex_to_base64(memo.memo_hash.hex()), Transaction.MEMO_TYPES.hash
    elif isinstance(memo, TextMemo):
        return memo.memo_text.decode(), Transaction.MEMO_TYPES.text
    else:
        raise ValueError()


def make_memo(
    memo: Optional[str], memo_type: Optional[str]
) -> Optional[Union[TextMemo, HashMemo, IdMemo]]:
    if not (memo or memo_type):
        return None
    if memo_type == Transaction.MEMO_TYPES.id:
        return IdMemo(int(memo))
    elif memo_type == Transaction.MEMO_TYPES.hash:
        return HashMemo(memo_base64_to_hex(memo))
    elif memo_type == Transaction.MEMO_TYPES.text:
        return TextMemo(memo)
    else:
        raise ValueError()


def validate_account_and_memo(account: str, memo: str):
    if not (isinstance(account, str) and isinstance(memo, str)):
        raise ValueError("invalid public key or memo type, expected strings")
    try:
        Keypair.from_public_key(account)
    except Ed25519PublicKeyInvalidError:
        raise ValueError("invalid public key")
    try:
        IdMemo(int(memo))
    except (ValueError, MemoInvalidException):
        pass
    else:
        return account, memo, Transaction.MEMO_TYPES.id
    try:
        HashMemo(memo_base64_to_hex(memo))
    except (ValueError, MemoInvalidException):
        pass
    else:
        return account, memo, Transaction.MEMO_TYPES.hash
    try:
        TextMemo(memo)
    except MemoInvalidException:
        raise ValueError("invalid memo")
    else:
        return account, memo, Transaction.MEMO_TYPES.text


SEP_9_FIELDS = {
    "family_name",
    "last_name",
    "given_name",
    "first_name",
    "additional_name",
    "address_country_code",
    "state_or_province",
    "city",
    "postal_code",
    "address",
    "mobile_number",
    "email_address",
    "birth_date",
    "birth_place",
    "birth_country_code",
    "bank_account_number",
    "bank_account_type",
    "bank_number",
    "bank_phone_number",
    "bank_branch_number",
    "tax_id",
    "tax_id_name",
    "occupation",
    "employer_name",
    "employer_address",
    "language_code",
    "id_type",
    "id_country_code",
    "id_issue_date",
    "id_expiration_date",
    "id_number",
    "photo_id_front",
    "photo_id_back",
    "notary_approval_of_photo_id",
    "ip_address",
    "photo_proof_residence",
    "sex",
    "proof_of_income",
    "proof_of_liveness",
    "organization.name",
    "organization.VAT_number",
    "organization.registration_number",
    "organization.registered_address",
    "organization.number_of_shareholders",
    "organization.shareholder_name",
    "organization.photo_incorporation_doc",
    "organization.photo_proof_address",
    "organization.address_country_code",
    "organization.state_or_province",
    "organization.city",
    "organization.postal_code",
    "organization.director_name",
    "organization.website",
    "organization.email",
    "organization.phone",
}


def extract_sep9_fields(args):
    sep9_args = {}
    for field in SEP_9_FIELDS:
        if field in args:
            sep9_args[field] = args.get(field)
    return sep9_args

def compute_callback_signature(callback_url: str, callback_body: str) -> str:
    callback_time = int(time.time())
    sig_payload = f"{callback_time}.{urlparse(callback_url).netloc}.{callback_body}"
    signature = base64.b64encode(Keypair.from_secret(settings.SIGNING_SEED).sign(sig_payload.encode())).decode()
    return f"t={callback_time}, s={signature}"

def make_on_change_callback(
    transaction: Transaction, timeout: Optional[int] = None
) -> Optional[RequestsResponse]:
    """
    Makes a POST request to `transaction.on_change_callback`, a URL
    provided by the client. The request will time out in
    ``settings.CALLBACK_REQUEST_TIMEOUT`` seconds if _timeout_ is not specified.

    The client is responsible for providing a publicly-accessible URL that
    responds within the timeout period. Polaris will continue processing
    `transaction` regardless of the result of this request.

    :raises: A ``requests.RequestException`` subclass or ``ValueError``
    :returns: The ``requests.Response`` object for the request
    """
    if (
        not transaction.on_change_callback
        or transaction.on_change_callback.lower() == "postmessage"
    ):
        raise ValueError("invalid or missing on_change_callback")
    if not timeout:
        timeout = settings.CALLBACK_REQUEST_TIMEOUT
    callback_body = {"transaction": TransactionSerializer(transaction).data}
    try:
        signature_header_value = compute_callback_signature(transaction.on_change_callback, callback_body)
    except ValueError:  # 
        logger.error(f"unable to parse host of transaction.on_change_callback for transaction {transaction.id}")
        return None
    headers = {"Signature": signature_header_value}
    return post(
        url=transaction.on_change_callback,
        json=callback_body,
        timeout=timeout,
        headers=headers
    )

def maybe_make_callback(transaction: Transaction, timeout: Optional[int] = None):
    """
    Makes the on_change_callback request if present on the transaciton and
    potentially logs an error. Use this function only if the response to the
    callback is irrelevant for your use case.
    """
    if (
        transaction.on_change_callback
        and transaction.on_change_callback.lower() != "postmessage"
    ):
        try:
            callback_resp = make_on_change_callback(transaction, timeout=timeout)
        except RequestException as e:
            logger.error(f"Callback request raised {e.__class__.__name__}: {str(e)}")
        else:
            if callback_resp and not callback_resp.ok:
                logger.error(f"Callback request returned {callback_resp.status_code}")


async def make_on_change_callback_async(
    transaction: Transaction, timeout: Optional[int] = None
) -> Optional[ClientResponse]:
    if (
        not transaction.on_change_callback
        or transaction.on_change_callback.lower() == "postmessage"
    ):
        raise ValueError("invalid or missing on_change_callback")
    if not timeout:
        timeout = settings.CALLBACK_REQUEST_TIMEOUT
    timeout_obj = aiohttp.ClientTimeout(total=timeout)
    callback_body = {"transaction": TransactionSerializer(transaction).data}
    try:
        signature_header_value = compute_callback_signature(transaction.on_change_callback, callback_body)
    except ValueError:  # 
        logger.error(f"unable to parse host of transaction.on_change_callback for transaction {transaction.id}")
        return None
    headers = {"Signature": signature_header_value}
    async with aiohttp.ClientSession(timeout=timeout_obj) as session:
        return await session.post(
            url=transaction.on_change_callback,
            json=callback_body,
            timeout=timeout,
            headers=headers
        )


async def maybe_make_callback_async(
    transaction: Transaction, timeout: Optional[int] = None
):
    """
    Makes the on_change_callback request if present on the transaciton and
    potentially logs an error. Use this function only if the response to the
    callback is irrelevant for your use case.
    """
    if (
        transaction.on_change_callback
        and transaction.on_change_callback.lower() != "postmessage"
    ):
        try:
            callback_resp = await make_on_change_callback_async(
                transaction, timeout=timeout
            )
        except RequestException as e:
            logger.error(f"Callback request raised {e.__class__.__name__}: {str(e)}")
        else:
            if callback_resp and not callback_resp.ok:
                logger.error(f"Callback request returned {callback_resp.status}")


def validate_patch_request_fields(fields: Dict, transaction: Transaction):
    try:
        required_info_updates = json.loads(transaction.required_info_updates)
    except (ValueError, TypeError):
        raise RuntimeError(
            "expected json-encoded string from transaction.required_info_update"
        )
    for category, expected_fields in required_info_updates.items():
        if category not in fields:
            raise ValueError(gettext("missing %s fields") % category)
        elif not isinstance(fields[category], dict):
            raise ValueError(
                gettext("invalid type for %s, must be an object") % category
            )
        for field in expected_fields:
            if field not in fields[category]:
                raise ValueError(
                    gettext("missing %(field)s in %(category)s")
                    % {"field": field, "category": category}
                )


def create_deposit_envelope(
    transaction, source_account, use_claimable_balance, base_fee
) -> TransactionEnvelope:
    if transaction.amount_out:
        payment_amount = transaction.amount_out
    elif transaction.quote:
        raise RuntimeError(
            f"transaction {transaction.id} uses a quote but does not have "
            "amount_out assigned"
        )
    else:
        payment_amount = round(
            Decimal(transaction.amount_in) - Decimal(transaction.amount_fee),
            transaction.asset.significant_decimals,
        )
    builder = TransactionBuilder(
        source_account=source_account,
        network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
        base_fee=base_fee,
    )
    asset = StellarAsset(code=transaction.asset.code, issuer=transaction.asset.issuer)
    if use_claimable_balance:
        claimant = Claimant(destination=transaction.to_address)
        builder.append_create_claimable_balance_op(
            claimants=[claimant],
            asset=asset,
            amount=str(payment_amount),
            source=transaction.asset.distribution_account,
        )
    else:
        builder.append_payment_op(
            destination=transaction.to_address,
            asset=asset,
            amount=str(payment_amount),
            source=transaction.asset.distribution_account,
        )
    if transaction.memo:
        builder.add_memo(make_memo(transaction.memo, transaction.memo_type))
    return builder.build()


def get_quote_and_offchain_destination_asset(
    token: SEP10Token,
    quote_id: str,
    destination_asset_str: str,
    asset: Asset,
    amount: Decimal,
) -> Tuple[Optional[Quote], Optional[OffChainAsset]]:
    quote = None
    destination_asset = None
    if quote_id:
        if "sep-38" not in settings.ACTIVE_SEPS or not asset.sep38_enabled:
            raise ValueError(_("quotes are not supported"))
        if not destination_asset_str:
            raise ValueError(
                _("'destination_asset' must be provided if 'quote_id' is provided")
            )
        try:
            quote = Quote.objects.get(
                id=quote_id,
                stellar_account=token.account,
                account_memo=token.memo,
                muxed_account=token.muxed_account,
                type=Quote.TYPE.firm,
            )
        except ObjectDoesNotExist:
            raise ValueError(_("quote not found"))
        if quote.expires_at < datetime.now(timezone.utc):
            raise ValueError(_("quote has expired"))
        if quote.sell_asset != asset.asset_identification_format:
            raise ValueError(
                _(
                    "quote 'sell_asset' does not match 'asset_code' and 'asset_issuer' parameters"
                )
            )
        if quote.buy_asset != destination_asset_str:
            raise ValueError(
                _("quote 'buy_asset' does not match 'destination_asset' parameter")
            )
        if quote.sell_amount != amount:
            raise ValueError(_("quote amount does not match 'amount' parameter"))
        try:
            destination_asset = OffChainAsset.objects.get(
                **asset_id_to_kwargs(destination_asset_str)
            )
        except (ValueError, TypeError, ObjectDoesNotExist):
            raise ValueError(_("invalid 'destination_asset'"))
    elif destination_asset_str:
        if "sep-38" not in settings.ACTIVE_SEPS or not asset.sep38_enabled:
            raise ValueError(_("quotes are not supported"))
        if not ExchangePair.objects.filter(
            sell_asset=asset.asset_identification_format,
            buy_asset=destination_asset_str,
        ).exists():
            raise ValueError(
                _("unsupported 'destination_asset' for 'asset_code' and 'asset_issuer'")
            )
        try:
            destination_asset = OffChainAsset.objects.get(
                **asset_id_to_kwargs(destination_asset_str)
            )
        except (ValueError, TypeError, ObjectDoesNotExist):
            raise ValueError(_("invalid 'destination_asset'"))
        quote = Quote(
            id=str(uuid.uuid4()),
            type=Quote.TYPE.indicative,
            stellar_account=token.account,
            account_memo=token.memo,
            muxed_account=token.muxed_account,
            sell_asset=asset.asset_identification_format,
            buy_asset=destination_asset_str,
            sell_amount=amount,
        )
    return quote, destination_asset


def get_quote_and_offchain_source_asset(
    token: SEP10Token,
    quote_id: str,
    source_asset_str: str,
    asset: Asset,
    amount: Decimal,
) -> Tuple[Optional[Quote], Optional[OffChainAsset]]:
    quote = None
    source_asset = None
    if quote_id:
        if "sep-38" not in settings.ACTIVE_SEPS or not asset.sep38_enabled:
            raise ValueError(_("quotes are not supported"))
        if not source_asset_str:
            raise ValueError(
                _("'source_asset' must be provided if 'quote_id' is provided")
            )
        try:
            quote = Quote.objects.get(
                id=quote_id,
                stellar_account=token.account,
                account_memo=token.memo,
                muxed_account=token.muxed_account,
                type=Quote.TYPE.firm,
            )
        except ObjectDoesNotExist:
            raise ValueError(_("quote not found"))
        if quote.expires_at < datetime.now(timezone.utc):
            raise ValueError(_("quote has expired"))
        if quote.buy_asset != asset.asset_identification_format:
            raise ValueError(
                _(
                    "quote 'buy_asset' does not match 'asset_code' and 'asset_issuer' parameters"
                )
            )
        if quote.sell_asset != source_asset_str:
            raise ValueError(
                _("quote 'sell_asset' does not match 'source_asset' parameter")
            )
        if quote.sell_amount != amount:
            raise ValueError(_("quote amount does not match 'amount' parameter"))
        try:
            source_asset = OffChainAsset.objects.get(
                **asset_id_to_kwargs(source_asset_str)
            )
        except (ValueError, TypeError, ObjectDoesNotExist):
            raise ValueError(_("invalid 'source_asset'"))
    elif source_asset_str:
        if "sep-38" not in settings.ACTIVE_SEPS or not asset.sep38_enabled:
            raise ValueError(_("quotes are not supported"))
        if not ExchangePair.objects.filter(
            sell_asset=source_asset_str, buy_asset=asset.asset_identification_format
        ).exists():
            raise ValueError(
                _("unsupported 'source_asset' for 'asset_code' and 'asset_issuer'")
            )
        try:
            source_asset = OffChainAsset.objects.get(
                **asset_id_to_kwargs(source_asset_str)
            )
        except (ValueError, TypeError, ObjectDoesNotExist):
            raise ValueError(_("invalid 'source_asset'"))
        quote = Quote(
            id=str(uuid.uuid4()),
            type=Quote.TYPE.indicative,
            stellar_account=token.account,
            account_memo=token.memo,
            muxed_account=token.muxed_account,
            buy_asset=asset.asset_identification_format,
            sell_asset=source_asset_str,
            sell_amount=amount,
        )
    return quote, source_asset



"""app URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog', include('blog.urls'))
"""
from django.urls import path, include, re_path

from polaris import settings


urlpatterns = []
if "sep-1" in settings.ACTIVE_SEPS:
    urlpatterns.append(path(".well-known/", include("polaris.sep1.urls")))

if "sep-6" in settings.ACTIVE_SEPS:
    urlpatterns.append(path("sep6/", include("polaris.sep6.urls")))

if "sep-10" in settings.ACTIVE_SEPS:
    urlpatterns.append(re_path(r"^auth/?", include("polaris.sep10.urls")))

if "sep-12" in settings.ACTIVE_SEPS:
    urlpatterns.append(path("kyc/", include("polaris.sep12.urls")))

if "sep-24" in settings.ACTIVE_SEPS:
    urlpatterns.append(path("sep24/", include("polaris.sep24.urls")))

if "sep-31" in settings.ACTIVE_SEPS:
    urlpatterns.append(path("sep31/", include("polaris.sep31.urls")))

if "sep-38" in settings.ACTIVE_SEPS:
    urlpatterns.append(path("sep38/", include("polaris.sep38.urls")))



# Generated by Django 2.2.16 on 2020-11-20 15:41

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0003_auto_20201009_1300"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="claimable_balance_id",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="transaction",
            name="claimable_balance_supported",
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.19 on 2021-03-17 23:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0007_transaction_on_change_callback"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="pending_execution_attempt",
            field=models.BooleanField(default=False),
        ),
    ]



from django.urls import re_path

from polaris import settings
from polaris.sep6 import info, deposit, withdraw, fee, transaction

urlpatterns = [
    re_path(r"^deposit/?$", deposit.deposit),
    re_path(r"^withdraw/?$", withdraw.withdraw),
    re_path(r"^info/?$", info.info),
    re_path(r"^fee/?$", fee.fee),
    re_path(r"^transaction/?$", transaction.transaction),
    re_path(
        r"^transactions/(?P<transaction_id>[^/]+)/?$", transaction.patch_transaction
    ),
    re_path(r"^transactions/?$", transaction.transactions),
]
if settings.SEP6_USE_MORE_INFO_URL:
    urlpatterns.append(
        re_path(
            r"^transaction/more_info/?$", transaction.more_info, name="more_info_sep6"
        ),
    )
if "sep-38" in settings.ACTIVE_SEPS:
    urlpatterns.extend(
        [
            re_path(r"^deposit-exchange/?$", deposit.deposit_exchange),
            re_path(r"^withdraw-exchange/?$", withdraw.withdraw_exchange),
        ]
    )
